@@ -26,7 +26,7 @@
 #include "erl_vm.h"
 #include "global.h"
 #include "erl_process.h"
-#include "beam_load.h"
+#include "beam_code.h"
 #include "bif.h"
 #include "error.h"
 #include "erl_binary.h"
@@ -34,6 +34,10 @@
 #include "erl_term.h"
 #include "erl_nfunc_sched.h"
 
+#include "beam_common.h"
+#include "jit/beam_asm.h"
+
+
 /* *************************************************************************
 ** Macros
 */
@@ -46,17 +50,6 @@
 #define ReAlloc(P, SIZ)		erts_realloc(ERTS_ALC_T_BPD, (P), (SZ))
 #define Free(P)			erts_free(ERTS_ALC_T_BPD, (P))
 
-#if defined(ERTS_ENABLE_LOCK_CHECK)
-#  define ERTS_REQ_PROC_MAIN_LOCK(P) \
-      if ((P)) erts_proc_lc_require_lock((P), ERTS_PROC_LOCK_MAIN,\
-					 __FILE__, __LINE__)
-#  define ERTS_UNREQ_PROC_MAIN_LOCK(P) \
-      if ((P)) erts_proc_lc_unrequire_lock((P), ERTS_PROC_LOCK_MAIN)
-#else
-#  define ERTS_REQ_PROC_MAIN_LOCK(P)
-#  define ERTS_UNREQ_PROC_MAIN_LOCK(P)
-#endif
-
 #define ERTS_BPF_LOCAL_TRACE       0x01
 #define ERTS_BPF_META_TRACE        0x02
 #define ERTS_BPF_COUNT             0x04
@@ -68,11 +61,6 @@
 
 #define ERTS_BPF_ALL               0xFF
 
-extern BeamInstr beam_return_to_trace[1];   /* OpCode(i_return_to_trace) */
-extern BeamInstr beam_return_trace[1];      /* OpCode(i_return_trace) */
-extern BeamInstr beam_exception_trace[1];   /* OpCode(i_exception_trace) */
-extern BeamInstr beam_return_time_trace[1]; /* OpCode(i_return_time_trace) */
-
 erts_atomic32_t erts_active_bp_index;
 erts_atomic32_t erts_staging_bp_index;
 erts_mtx_t erts_dirty_bp_ix_mtx;
@@ -126,10 +114,10 @@ static void set_function_break(ErtsCodeInfo *ci,
 			       ErtsTracer tracer);
 
 static void clear_break(BpFunctions* f, Uint break_flags);
-static int clear_function_break(ErtsCodeInfo *ci, Uint break_flags);
+static int clear_function_break(const ErtsCodeInfo *ci, Uint break_flags);
 
-static BpDataTime* get_time_break(ErtsCodeInfo *ci);
-static GenericBpData* check_break(ErtsCodeInfo *ci, Uint break_flags);
+static BpDataTime* get_time_break(const ErtsCodeInfo *ci);
+static GenericBpData* check_break(const ErtsCodeInfo *ci, Uint break_flags);
 
 static void bp_meta_unref(BpMetaTracer *bmt);
 static void bp_count_unref(BpCount *bcp);
@@ -188,8 +176,7 @@ erts_bp_match_functions(BpFunctions* f, ErtsCodeMFA *mfa, int specified)
     f->matching = (BpFunction *) Alloc(max_funcs*sizeof(BpFunction));
     i = 0;
     for (current = 0; current < num_modules; current++) {
-	BeamCodeHeader* code_hdr = module[current]->curr.code_hdr;
-	ErtsCodeInfo* ci;
+        const BeamCodeHeader* code_hdr = module[current]->curr.code_hdr;
 	Uint num_functions = (Uint)(UWord) code_hdr->num_functions;
 	Uint fi;
 
@@ -201,12 +188,16 @@ erts_bp_match_functions(BpFunctions* f, ErtsCodeMFA *mfa, int specified)
 	}
 
 	for (fi = 0; fi < num_functions; fi++) {
+            ErtsCodeInfo* ci;
+            void *w_ptr;
 
-	    ci = code_hdr->functions[fi];
+            w_ptr = erts_writable_code_ptr(&module[current]->curr,
+                                           code_hdr->functions[fi]);
+            ci = (ErtsCodeInfo*)w_ptr;
+
+#ifndef BEAMASM
 	    ASSERT(BeamIsOpCode(ci->op, op_i_func_info_IaaI));
-	    if (erts_is_function_native(ci)) {
-		continue;
-	    }
+#endif
             switch (specified) {
             case 3:
                 if (ci->mfa.arity != mfa->arity)
@@ -241,7 +232,6 @@ erts_bp_match_export(BpFunctions* f, ErtsCodeMFA *mfa, int specified)
     f->matching = (BpFunction *) Alloc(num_exps*sizeof(BpFunction));
     ne = 0;
     for (i = 0; i < num_exps; i++) {
-        BeamInstr *func;
         Export* ep;
 
         ep = export_list(i, code_ix);
@@ -262,15 +252,12 @@ erts_bp_match_export(BpFunctions* f, ErtsCodeMFA *mfa, int specified)
             ASSERT(0);
         }
 
-        func = ep->addressv[code_ix];
-
-        if (func == ep->trampoline.raw) {
-            if (BeamIsOpCode(*func, op_call_error_handler)) {
-                    continue;
+        if (erts_is_export_trampoline_active(ep, code_ix)) {
+            if (BeamIsOpCode(ep->trampoline.common.op, op_call_error_handler)) {
+                continue;
             }
-            ASSERT(BeamIsOpCode(*func, op_i_generic_breakpoint));
-        } else if (erts_is_function_native(erts_code_to_codeinfo(func))) {
-            continue;
+
+            ASSERT(BeamIsOpCode(ep->trampoline.common.op, op_i_generic_breakpoint));
         }
 
 	f->matching[ne].ci = &ep->info;
@@ -354,8 +341,12 @@ consolidate_bp_data(Module* modp, ErtsCodeInfo *ci, int local)
 	    }
 	    ASSERT(modp->curr.num_breakpoints >= 0);
 	    ASSERT(modp->curr.num_traced_exports >= 0);
-	    ASSERT(! BeamIsOpCode(*erts_codeinfo_to_code(ci),
-                                  op_i_generic_breakpoint));
+#if !defined(BEAMASM) && defined(DEBUG)
+            {
+                BeamInstr instr = *(const BeamInstr*)erts_codeinfo_to_code(ci);
+                ASSERT(!BeamIsOpCode(instr, op_i_generic_breakpoint));
+            }
+#endif
 	}
 	ci->u.gen_bp = NULL;
 	Free(g);
@@ -403,16 +394,29 @@ erts_install_breakpoints(BpFunctions* f)
 {
     Uint i;
     Uint n = f->matched;
-    BeamInstr br = BeamOpCodeAddr(op_i_generic_breakpoint);
 
     for (i = 0; i < n; i++) {
 	ErtsCodeInfo* ci = f->matching[i].ci;
 	GenericBp* g = ci->u.gen_bp;
-        BeamInstr volatile *pc = erts_codeinfo_to_code(ci);
-        BeamInstr instr = *pc;
+        Module* modp = f->matching[i].mod;
+#ifdef BEAMASM
+        if ((erts_asm_bp_get_flags(ci) & ERTS_ASM_BP_FLAG_BP) == 0 && g) {
+	    /*
+	     * The breakpoint must be disabled in the active data
+	     * (it will enabled later by switching bp indices),
+	     * and enabled in the staging data.
+	     */
+	    ASSERT(g->data[erts_active_bp_ix()].flags == 0);
+	    ASSERT(g->data[erts_staging_bp_ix()].flags != 0);
 
+            erts_asm_bp_set_flag(ci, ERTS_ASM_BP_FLAG_BP);
+            modp->curr.num_breakpoints++;
+        }
+#else
+        BeamInstr volatile *pc = (BeamInstr*)erts_codeinfo_to_code(ci);
+        BeamInstr instr = *pc;
 	if (!BeamIsOpCode(instr, op_i_generic_breakpoint) && g) {
-	    Module* modp = f->matching[i].mod;
+            BeamInstr br = BeamOpCodeAddr(op_i_generic_breakpoint);
 
 	    /*
 	     * The breakpoint must be disabled in the active data
@@ -437,6 +441,7 @@ erts_install_breakpoints(BpFunctions* f)
             *pc = instr;
 	    modp->curr.num_breakpoints++;
 	}
+#endif
     }
 }
 
@@ -447,28 +452,44 @@ erts_uninstall_breakpoints(BpFunctions* f)
     Uint n = f->matched;
 
     for (i = 0; i < n; i++) {
-	uninstall_breakpoint(f->matching[i].ci);
+        uninstall_breakpoint(f->matching[i].ci);
     }
 }
 
+#ifdef BEAMASM
 static void
 uninstall_breakpoint(ErtsCodeInfo *ci)
 {
-    BeamInstr *pc = erts_codeinfo_to_code(ci);
-    if (BeamIsOpCode(*pc, op_i_generic_breakpoint)) {
-	GenericBp* g = ci->u.gen_bp;
-	if (g->data[erts_active_bp_ix()].flags == 0) {
-	    /*
-	     * The following write is not protected by any lock. We
-	     * assume that the hardware guarantees that a write of an
-	     * aligned word-size (or half-word) writes is atomic
-	     * (i.e. that other processes executing this code will not
-	     * see a half pointer).
-	     */
-	    *pc = g->orig_instr;
-	}
+    if (erts_asm_bp_get_flags(ci) & ERTS_ASM_BP_FLAG_BP) {
+        GenericBp* g = ci->u.gen_bp;
+
+        if (g->data[erts_active_bp_ix()].flags == 0) {
+            erts_asm_bp_unset_flag(ci, ERTS_ASM_BP_FLAG_BP);
+        }
     }
 }
+#else
+static void
+uninstall_breakpoint(ErtsCodeInfo *ci)
+{
+    BeamInstr *pc = (BeamInstr*)erts_codeinfo_to_code(ci);
+
+    if (BeamIsOpCode(*pc, op_i_generic_breakpoint)) {
+        GenericBp* g = ci->u.gen_bp;
+
+        if (g->data[erts_active_bp_ix()].flags == 0) {
+            /*
+             * The following write is not protected by any lock. We
+             * assume that the hardware guarantees that a write of an
+             * aligned word-size (or half-word) writes is atomic
+             * (i.e. that other processes executing this code will not
+             * see a half pointer).
+             */
+            *pc = g->orig_instr;
+        }
+    }
+}
+#endif
 
 void
 erts_set_trace_break(BpFunctions* f, Binary *match_spec)
@@ -521,10 +542,11 @@ erts_clear_export_trace(ErtsCodeInfo *ci, int local)
     GenericBp* g = ci->u.gen_bp;
 
     if (g) {
-	Uint flags = local ? ERTS_BPF_LOCAL_TRACE : ERTS_BPF_GLOBAL_TRACE;
-	if (g->data[erts_staging_bp_ix()].flags & flags) {
-	    clear_function_break(ci, flags);
-	}
+        Uint flags = local ? ERTS_BPF_LOCAL_TRACE : ERTS_BPF_GLOBAL_TRACE;
+
+        if (g->data[erts_staging_bp_ix()].flags & flags) {
+            clear_function_break(ci, flags);
+        }
     }
 }
 
@@ -559,36 +581,42 @@ erts_clear_all_breaks(BpFunctions* f)
     clear_break(f, ERTS_BPF_ALL);
 }
 
-int 
+int
 erts_clear_module_break(Module *modp) {
-    BeamCodeHeader* code_hdr;
+    const BeamCodeHeader* code_hdr;
     Uint n;
     Uint i;
 
     ERTS_LC_ASSERT(erts_thr_progress_is_blocking());
     ASSERT(modp);
+
     code_hdr = modp->curr.code_hdr;
     if (!code_hdr) {
-	return 0;
+        return 0;
     }
-    n = (Uint)(UWord) code_hdr->num_functions;
+
+    n = (Uint)code_hdr->num_functions;
     for (i = 0; i < n; ++i) {
-	ErtsCodeInfo *ci = code_hdr->functions[i];
-	if (erts_is_function_native(ci))
-	    continue;
-	clear_function_break(ci, ERTS_BPF_ALL);
+        const ErtsCodeInfo *ci = code_hdr->functions[i];
+
+        clear_function_break(ci, ERTS_BPF_ALL);
     }
 
     erts_commit_staged_bp();
 
     for (i = 0; i < n; ++i) {
-	ErtsCodeInfo *ci = code_hdr->functions[i];
-	if (erts_is_function_native(ci))
-	    continue;
-	uninstall_breakpoint(ci);
-	consolidate_bp_data(modp, ci, 1);
-	ASSERT(ci->u.gen_bp == NULL);
+        ErtsCodeInfo* ci;
+        void *w_ptr;
+
+        w_ptr = erts_writable_code_ptr(&modp->curr,
+                                       code_hdr->functions[i]);
+        ci = (ErtsCodeInfo*)w_ptr;
+
+        uninstall_breakpoint(ci);
+        consolidate_bp_data(modp, ci, 1);
+        ASSERT(ci->u.gen_bp == NULL);
     }
+
     return n;
 }
 
@@ -602,9 +630,10 @@ erts_clear_export_break(Module* modp, Export *ep)
     ci = &ep->info;
 
     ASSERT(erts_codeinfo_to_code(ci) == ep->trampoline.raw);
-
-    ASSERT(BeamIsOpCode(ep->trampoline.op, op_i_generic_breakpoint));
-    ep->trampoline.op = 0;
+#ifndef BEAMASM
+    ASSERT(BeamIsOpCode(ep->trampoline.common.op, op_i_generic_breakpoint));
+#endif
+    ep->trampoline.common.op = 0;
 
     clear_function_break(ci, ERTS_BPF_ALL);
     erts_commit_staged_bp();
@@ -629,13 +658,13 @@ static void fixup_cp_before_trace(Process *c_p, int *return_to_trace)
     Eterm *cpp = c_p->stop;
 
     for (;;) {
-        BeamInstr w = *cp_val(*cpp);
-        if (BeamIsOpCode(w, op_return_trace)) {
+        ErtsCodePtr w = cp_val(*cpp);
+        if (BeamIsReturnTrace(w)) {
             cpp += 3;
-        } else if (BeamIsOpCode(w, op_i_return_to_trace)) {
+        } else if (BeamIsReturnToTrace(w)) {
             *return_to_trace = 1;
             cpp += 1;
-        } else if (BeamIsOpCode(w, op_i_return_time_trace)) {
+        } else if (BeamIsReturnTimeTrace(w)) {
             cpp += 2;
         } else {
             break;
@@ -653,7 +682,9 @@ erts_generic_breakpoint(Process* c_p, ErtsCodeInfo *info, Eterm* reg)
     Uint bp_flags;
     ErtsBpIndex ix = erts_active_bp_ix();
 
+#ifndef BEAMASM
     ASSERT(BeamIsOpCode(info->op, op_i_func_info_IaaI));
+#endif
 
     g = info->u.gen_bp;
     bp = &g->data[ix];
@@ -687,7 +718,7 @@ erts_generic_breakpoint(Process* c_p, ErtsCodeInfo *info, Eterm* reg)
 	new_tracer = do_call_trace(c_p, info, reg, 1, bp->meta_ms, old_tracer);
 
 	if (!ERTS_TRACER_COMPARE(new_tracer, old_tracer)) {
-            if (old_tracer == erts_atomic_cmpxchg_acqb(
+            if ((erts_aint_t)old_tracer == erts_atomic_cmpxchg_acqb(
                     &bp->meta_tracer->tracer,
                     (erts_aint_t)new_tracer,
                     (erts_aint_t)old_tracer)) {
@@ -703,16 +734,16 @@ erts_generic_breakpoint(Process* c_p, ErtsCodeInfo *info, Eterm* reg)
     }
 
     if (bp_flags & ERTS_BPF_TIME_TRACE_ACTIVE) {
-	BeamInstr w;
+        const ErtsCodeInfo* prev_info;
+        ErtsCodePtr w;
         Eterm* E;
-	ErtsCodeInfo* prev_info = erts_trace_time_call(c_p, info, bp->time);
+
+        prev_info= erts_trace_time_call(c_p, info, bp->time);
         E = c_p->stop;
-        w = *(BeamInstr*) E[0];
-	if (! (BeamIsOpCode(w, op_i_return_time_trace) ||
-	       BeamIsOpCode(w, op_return_trace) ||
-               BeamIsOpCode(w, op_i_return_to_trace)) ) {
+        w = (ErtsCodePtr) E[0];
+	if (!(BeamIsReturnTrace(w) || BeamIsReturnToTrace(w) || BeamIsReturnTimeTrace(w))) {
 	    ASSERT(c_p->htop <= E && E <= c_p->hend);
-	    if (E - 2 < c_p->htop) {
+	    if (HeapWordsLeft(c_p) < 2) {
 		(void) erts_garbage_collect(c_p, 2, reg, info->mfa.arity);
 		ERTS_VERIFY_UNUSED_TEMP_ALLOC(c_p);
 	    }
@@ -728,7 +759,7 @@ erts_generic_breakpoint(Process* c_p, ErtsCodeInfo *info, Eterm* reg)
     }
 
     if (bp_flags & ERTS_BPF_DEBUG) {
-	return BeamOpCodeAddr(op_i_debug_breakpoint);
+        return BeamOpCodeAddr(op_i_debug_breakpoint);
     } else {
 	return g->orig_instr;
     }
@@ -762,7 +793,7 @@ do_call_trace(Process* c_p, ErtsCodeInfo* info, Eterm* reg,
     }
     if (need) {
 	ASSERT(c_p->htop <= E && E <= c_p->hend);
-	if (E - need < c_p->htop) {
+	if (HeapWordsLeft(c_p) < need) {
 	    (void) erts_garbage_collect(c_p, need, reg, info->mfa.arity);
 	    ERTS_VERIFY_UNUSED_TEMP_ALLOC(c_p);
 	    E = c_p->stop;
@@ -791,8 +822,8 @@ do_call_trace(Process* c_p, ErtsCodeInfo* info, Eterm* reg,
     return tracer;
 }
 
-ErtsCodeInfo*
-erts_trace_time_call(Process* c_p, ErtsCodeInfo *info, BpDataTime* bdt)
+const ErtsCodeInfo*
+erts_trace_time_call(Process* c_p, const ErtsCodeInfo *info, BpDataTime* bdt)
 {
     ErtsMonotonicTime time;
     process_breakpoint_time_t *pbt = NULL;
@@ -800,7 +831,7 @@ erts_trace_time_call(Process* c_p, ErtsCodeInfo *info, BpDataTime* bdt)
     bp_time_hash_t *h = NULL;
     BpDataTime *pbdt = NULL;
     Uint32 six = acquire_bp_sched_ix(c_p);
-    ErtsCodeInfo* prev_info;
+    const ErtsCodeInfo* prev_info;
 
     ASSERT(c_p);
     ASSERT(erts_atomic32_read_acqb(&c_p->state) & (ERTS_PSFLG_RUNNING
@@ -878,7 +909,7 @@ erts_trace_time_call(Process* c_p, ErtsCodeInfo *info, BpDataTime* bdt)
 }
 
 void
-erts_trace_time_return(Process *p, ErtsCodeInfo *prev_info)
+erts_trace_time_return(Process *p, const ErtsCodeInfo *prev_info)
 {
     ErtsMonotonicTime time;
     process_breakpoint_time_t *pbt = NULL;
@@ -945,7 +976,7 @@ erts_trace_time_return(Process *p, ErtsCodeInfo *prev_info)
 }
 
 int 
-erts_is_trace_break(ErtsCodeInfo *ci, Binary **match_spec_ret, int local)
+erts_is_trace_break(const ErtsCodeInfo *ci, Binary **match_spec_ret, int local)
 {
     Uint flags = local ? ERTS_BPF_LOCAL_TRACE : ERTS_BPF_GLOBAL_TRACE;
     GenericBpData* bp = check_break(ci, flags);
@@ -960,7 +991,7 @@ erts_is_trace_break(ErtsCodeInfo *ci, Binary **match_spec_ret, int local)
 }
 
 int
-erts_is_mtrace_break(ErtsCodeInfo *ci, Binary **match_spec_ret,
+erts_is_mtrace_break(const ErtsCodeInfo *ci, Binary **match_spec_ret,
 		     ErtsTracer *tracer_ret)
 {
     GenericBpData* bp = check_break(ci, ERTS_BPF_META_TRACE);
@@ -978,7 +1009,7 @@ erts_is_mtrace_break(ErtsCodeInfo *ci, Binary **match_spec_ret,
 }
 
 int 
-erts_is_count_break(ErtsCodeInfo *ci, Uint *count_ret)
+erts_is_count_break(const ErtsCodeInfo *ci, Uint *count_ret)
 {
     GenericBpData* bp = check_break(ci, ERTS_BPF_COUNT);
     
@@ -991,7 +1022,7 @@ erts_is_count_break(ErtsCodeInfo *ci, Uint *count_ret)
     return 0;
 }
 
-int erts_is_time_break(Process *p, ErtsCodeInfo *ci, Eterm *retval) {
+int erts_is_time_break(Process *p, const ErtsCodeInfo *ci, Eterm *retval) {
     Uint i, ix;
     bp_time_hash_t hash;
     Uint size;
@@ -1051,11 +1082,11 @@ int erts_is_time_break(Process *p, ErtsCodeInfo *ci, Eterm *retval) {
 }
 
 
-ErtsCodeInfo *
-erts_find_local_func(ErtsCodeMFA *mfa) {
+const ErtsCodeInfo *
+erts_find_local_func(const ErtsCodeMFA *mfa) {
+    const BeamCodeHeader *code_hdr;
+    const ErtsCodeInfo *ci;
     Module *modp;
-    BeamCodeHeader* code_hdr;
-    ErtsCodeInfo* ci;
     Uint i,n;
 
     if ((modp = erts_get_module(mfa->module, erts_active_code_ix())) == NULL)
@@ -1065,7 +1096,9 @@ erts_find_local_func(ErtsCodeMFA *mfa) {
     n = (BeamInstr) code_hdr->num_functions;
     for (i = 0; i < n; ++i) {
 	ci = code_hdr->functions[i];
+#ifndef BEAMASM
 	ASSERT(BeamIsOpCode(ci->op, op_i_func_info_IaaI));
+#endif
 	ASSERT(mfa->module == ci->mfa.module || is_nil(ci->mfa.module));
 	if (mfa->function == ci->mfa.function &&
 	    mfa->arity == ci->mfa.arity) {
@@ -1096,6 +1129,8 @@ static void bp_hash_rehash(bp_time_hash_t *hash, Uint n) {
     Uint ix;
     Uint hval;
 
+    ASSERT(n > 0);
+
     item = (bp_data_time_item_t *)Alloc(size);
     sys_memzero(item, size);
 
@@ -1103,6 +1138,7 @@ static void bp_hash_rehash(bp_time_hash_t *hash, Uint n) {
 	item[ix].pid = NIL;
     }
 
+
     /* rehash, old hash -> new hash */
 
     for( ix = 0; ix < hash->n; ix++) {
@@ -1275,7 +1311,23 @@ set_function_break(ErtsCodeInfo *ci, Binary *match_spec, Uint break_flags,
 	    return;
 	}
 	g = Alloc(sizeof(GenericBp));
-	g->orig_instr = *erts_codeinfo_to_code(ci);
+        {
+            const UWord *instr_word = (const UWord *)erts_codeinfo_to_code(ci);
+
+#ifdef BEAMASM
+            /* The orig_instr is only used in global tracing for BEAMASM and
+             * there the address i located within the trampoline in the export
+             * entry so we read it from there.
+             *
+             * For local tracing this value is patched in
+             * erts_set_trace_pattern. */
+            g->orig_instr = instr_word[2];
+#else
+            ERTS_CT_ASSERT(sizeof(UWord) == sizeof(BeamInstr));
+            g->orig_instr = instr_word[0];
+#endif
+        }
+
 	for (i = 0; i < ERTS_NUM_BP_IX; i++) {
 	    g->data[i].flags = 0;
 	}
@@ -1346,7 +1398,7 @@ set_function_break(ErtsCodeInfo *ci, Binary *match_spec, Uint break_flags,
 	bp->count = bcp;
     } else if (break_flags & ERTS_BPF_TIME_TRACE) {
 	BpDataTime* bdt;
-	int i;
+	Uint i;
 
 	ASSERT((bp->flags & ERTS_BPF_TIME_TRACE) == 0);
 	bdt = Alloc(sizeof(BpDataTime));
@@ -1376,7 +1428,7 @@ clear_break(BpFunctions* f, Uint break_flags)
 }
 
 static int
-clear_function_break(ErtsCodeInfo *ci, Uint break_flags)
+clear_function_break(const ErtsCodeInfo *ci, Uint break_flags)
 {
     GenericBp* g;
     GenericBpData* bp;
@@ -1446,27 +1498,29 @@ bp_time_unref(BpDataTime* bdt)
 }
 
 static BpDataTime*
-get_time_break(ErtsCodeInfo *ci)
+get_time_break(const ErtsCodeInfo *ci)
 {
     GenericBpData* bp = check_break(ci, ERTS_BPF_TIME_TRACE);
     return bp ? bp->time : 0;
 }
 
 static GenericBpData*
-check_break(ErtsCodeInfo *ci, Uint break_flags)
+check_break(const ErtsCodeInfo *ci, Uint break_flags)
 {
     GenericBp* g = ci->u.gen_bp;
 
+#ifndef BEAMASM
     ASSERT(BeamIsOpCode(ci->op, op_i_func_info_IaaI));
-    if (erts_is_function_native(ci)) {
-	return 0;
-    }
+#endif
+
     if (g) {
-	GenericBpData* bp = &g->data[erts_active_bp_ix()];
-	ASSERT((bp->flags & ~ERTS_BPF_ALL) == 0);
-	if (bp->flags & break_flags) {
-	    return bp;
-	}
+        GenericBpData* bp = &g->data[erts_active_bp_ix()];
+
+        ASSERT((bp->flags & ~ERTS_BPF_ALL) == 0);
+        if (bp->flags & break_flags) {
+            return bp;
+        }
     }
+
     return 0;
 }
