@@ -1,7 +1,7 @@
 /*
  * %CopyrightBegin%
  *
- * Copyright Ericsson AB 1996-2018. All Rights Reserved.
+ * Copyright Ericsson AB 1996-2021. All Rights Reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -21,6 +21,9 @@
 #ifndef __BIF_H__
 #define __BIF_H__
 
+#include "erl_process.h"
+#include "export.h"
+
 extern Export *erts_await_result;
 extern Export* erts_format_cpu_topology_trap;
 extern Export *erts_convert_time_unit_trap;
@@ -29,7 +32,7 @@ extern Export *erts_convert_time_unit_trap;
 
 #define BIF_P A__p
 
-#define BIF_ALIST Process* A__p, Eterm* BIF__ARGS, BeamInstr *A__I
+#define BIF_ALIST Process* A__p, Eterm* BIF__ARGS, ErtsCodePtr A__I
 #define BIF_CALL_ARGS A__p, BIF__ARGS, A__I
 
 #define BIF_ALIST_0 BIF_ALIST
@@ -58,6 +61,62 @@ extern Export *erts_convert_time_unit_trap;
 
 typedef BIF_RETTYPE (*ErtsBifFunc)(BIF_ALIST);
 
+#ifdef BEAMASM
+#define ERTS_REDS_LEFT(p, FCalls) (FCalls)
+
+#define ERTS_IS_PROC_OUT_OF_REDS(p)	((p)->fcalls == 0)
+
+#define BUMP_REDS(p, gc) do {			   \
+     ASSERT(p);		 			   \
+     ERTS_LC_ASSERT(ERTS_PROC_LOCK_MAIN & erts_proc_lc_my_proc_locks(p));\
+     (p)->fcalls -= (gc); 			   \
+     if ((p)->fcalls < 0) { 			   \
+           (p)->fcalls = 0; 			   \
+     } 						   \
+} while(0)
+
+#define BUMP_ALL_REDS(p) do {			\
+    (p)->fcalls = 0; 			\
+    ASSERT(ERTS_BIF_REDS_LEFT((p)) == 0);	\
+} while(0)
+
+#define ERTS_VBUMP_REDS(p, reds)					\
+do {									\
+	if ((p)->fcalls >= reds) {					\
+	    (p)->fcalls -= reds;					\
+	    erts_proc_sched_data((p))->virtual_reds += reds;		\
+	}								\
+	else {								\
+	    if ((p)->fcalls > 0)					\
+		erts_proc_sched_data((p))->virtual_reds += (p)->fcalls;	\
+	    (p)->fcalls = 0;						\
+	}								\
+} while(0)
+
+
+#define ERTS_VBUMP_ALL_REDS_INTERNAL(p, fcalls)				\
+do {									\
+	if ((fcalls) > 0)						\
+	    erts_proc_sched_data((p))->virtual_reds += (fcalls);	\
+	(fcalls) = 0;							\
+} while(0)
+
+#define ERTS_VBUMP_LEAVE_REDS_INTERNAL(P, Reds, FCalls)			\
+    do {								\
+	    if ((FCalls) > (Reds)) {					\
+		erts_proc_sched_data((P))->virtual_reds			\
+		    += (FCalls) - (Reds);				\
+		(FCalls) = (Reds);					\
+	    }								\
+    } while (0)
+
+#else
+
+#define ERTS_REDS_LEFT(p, FCalls)					\
+  (ERTS_PROC_GET_SAVED_CALLS_BUF((p))					\
+   ? ((FCalls) > -CONTEXT_REDS ? ((FCalls) - (-CONTEXT_REDS)) : 0)	\
+   : ((FCalls) > 0 ? (FCalls) : 0))
+
 #define ERTS_IS_PROC_OUT_OF_REDS(p)		\
     ((p)->fcalls > 0				\
      ? 0					\
@@ -65,32 +124,6 @@ typedef BIF_RETTYPE (*ErtsBifFunc)(BIF_ALIST);
 	? (p)->fcalls == 0			\
 	: ((p)->fcalls == -CONTEXT_REDS)))
 
-#define BUMP_ALL_REDS(p) do {			\
-    if (!ERTS_PROC_GET_SAVED_CALLS_BUF((p))) 	\
-	(p)->fcalls = 0; 			\
-    else 					\
-	(p)->fcalls = -CONTEXT_REDS;		\
-    ASSERT(ERTS_BIF_REDS_LEFT((p)) == 0);	\
-} while(0)
-
-#define ERTS_VBUMP_ALL_REDS_INTERNAL(p, fcalls)				\
-do {									\
-    if (!ERTS_PROC_GET_SAVED_CALLS_BUF((p))) {				\
-	if ((fcalls) > 0)						\
-	    erts_proc_sched_data((p))->virtual_reds += (fcalls);	\
-	(fcalls) = 0;							\
-    }									\
-    else {								\
-	if ((fcalls) > -CONTEXT_REDS)					\
-	    erts_proc_sched_data((p))->virtual_reds			\
-		+= ((fcalls) - (-CONTEXT_REDS));			\
-	(fcalls) = -CONTEXT_REDS;					\
-    }									\
-} while(0)
-
-#define ERTS_VBUMP_ALL_REDS(p) \
-    ERTS_VBUMP_ALL_REDS_INTERNAL((p), (p)->fcalls)
-
 #define BUMP_REDS(p, gc) do {			   \
      ASSERT(p);		 			   \
      ERTS_LC_ASSERT(ERTS_PROC_LOCK_MAIN & erts_proc_lc_my_proc_locks(p));\
@@ -103,6 +136,13 @@ do {									\
      } 						   \
 } while(0)
 
+#define BUMP_ALL_REDS(p) do {			\
+    if (!ERTS_PROC_GET_SAVED_CALLS_BUF((p))) 	\
+	(p)->fcalls = 0; 			\
+    else 					\
+	(p)->fcalls = -CONTEXT_REDS;		\
+    ASSERT(ERTS_BIF_REDS_LEFT((p)) == 0);	\
+} while(0)
 
 #define ERTS_VBUMP_REDS(p, reds)					\
 do {									\
@@ -131,6 +171,21 @@ do {									\
     }									\
 } while(0)
 
+#define ERTS_VBUMP_ALL_REDS_INTERNAL(p, fcalls)				\
+do {									\
+    if (!ERTS_PROC_GET_SAVED_CALLS_BUF((p))) {				\
+	if ((fcalls) > 0)						\
+	    erts_proc_sched_data((p))->virtual_reds += (fcalls);	\
+	(fcalls) = 0;							\
+    }									\
+    else {								\
+	if ((fcalls) > -CONTEXT_REDS)					\
+	    erts_proc_sched_data((p))->virtual_reds			\
+		+= ((fcalls) - (-CONTEXT_REDS));			\
+	(fcalls) = -CONTEXT_REDS;					\
+    }									\
+} while(0)
+
 #define ERTS_VBUMP_LEAVE_REDS_INTERNAL(P, Reds, FCalls)			\
     do {								\
 	if (ERTS_PROC_GET_SAVED_CALLS_BUF((P))) {			\
@@ -150,15 +205,16 @@ do {									\
 	}								\
     } while (0)
 
-#define ERTS_VBUMP_LEAVE_REDS(P, Reds)					\
+#endif
+
+#define ERTS_VBUMP_ALL_REDS(p) \
+    ERTS_VBUMP_ALL_REDS_INTERNAL((p), (p)->fcalls)
+
+#define ERTS_VBUMP_LEAVE_REDS(P, Reds) \
     ERTS_VBUMP_LEAVE_REDS_INTERNAL(P, Reds, (P)->fcalls)
 
-#define ERTS_REDS_LEFT(p, FCalls)					\
-  (ERTS_PROC_GET_SAVED_CALLS_BUF((p))					\
-   ? ((FCalls) > -CONTEXT_REDS ? ((FCalls) - (-CONTEXT_REDS)) : 0)	\
-   : ((FCalls) > 0 ? (FCalls) : 0))
-
-#define ERTS_BIF_REDS_LEFT(p) ERTS_REDS_LEFT(p, p->fcalls)
+#define ERTS_BIF_REDS_LEFT(p) \
+    ERTS_REDS_LEFT(p, p->fcalls)
 
 #define BIF_RET2(x, gc) do {			\
     BUMP_REDS(BIF_P, (gc));			\
@@ -183,7 +239,7 @@ do {								\
 
 #define ERTS_BIF_ERROR_TRAPPED1(Proc, Reason, Bif, A0)		\
 do {								\
-    Eterm* reg = erts_proc_sched_data((Proc))->x_reg_array;	\
+    Eterm* reg = erts_proc_sched_data((Proc))->registers->x_reg_array.d;	\
     (Proc)->freason = (Reason);					\
     (Proc)->current = &(Bif)->info.mfa;                         \
     reg[0] = (Eterm) (A0);					\
@@ -192,7 +248,7 @@ do {								\
 
 #define ERTS_BIF_ERROR_TRAPPED2(Proc, Reason, Bif, A0, A1)	\
 do {								\
-    Eterm* reg = erts_proc_sched_data((Proc))->x_reg_array;	\
+    Eterm* reg = erts_proc_sched_data((Proc))->registers->x_reg_array.d;	\
     (Proc)->freason = (Reason);					\
     (Proc)->current = &(Bif)->info.mfa;                         \
     reg[0] = (Eterm) (A0);					\
@@ -202,7 +258,7 @@ do {								\
 
 #define ERTS_BIF_ERROR_TRAPPED3(Proc, Reason, Bif, A0, A1, A2)	\
 do {								\
-    Eterm* reg = erts_proc_sched_data((Proc))->x_reg_array;	\
+    Eterm* reg = erts_proc_sched_data((Proc))->registers->x_reg_array.d;	\
     (Proc)->freason = (Reason);					\
     (Proc)->current = &(Bif)->info.mfa;                         \
     reg[0] = (Eterm) (A0);					\
@@ -226,7 +282,7 @@ do {								\
 
 #define ERTS_BIF_PREP_ERROR_TRAPPED1(Ret, Proc, Reason, Bif, A0) \
 do {								\
-    Eterm* reg = erts_proc_sched_data((Proc))->x_reg_array;	\
+    Eterm* reg = erts_proc_sched_data((Proc))->registers->x_reg_array.d;	\
     (Proc)->freason = (Reason);					\
     (Proc)->current = &(Bif)->info.mfa;                         \
     reg[0] = (Eterm) (A0);					\
@@ -235,7 +291,7 @@ do {								\
 
 #define ERTS_BIF_PREP_ERROR_TRAPPED2(Ret, Proc, Reason, Bif, A0, A1) \
 do {								\
-    Eterm* reg = erts_proc_sched_data((Proc))->x_reg_array;	\
+    Eterm* reg = erts_proc_sched_data((Proc))->registers->x_reg_array.d;	\
     (Proc)->freason = (Reason);					\
     (Proc)->current = &(Bif)->info.mfa;                         \
     reg[0] = (Eterm) (A0);					\
@@ -245,7 +301,7 @@ do {								\
 
 #define ERTS_BIF_PREP_ERROR_TRAPPED3(Ret, Proc, Reason, Bif, A0, A1, A2) \
 do {								\
-    Eterm* reg = erts_proc_sched_data((Proc))->x_reg_array;	\
+    Eterm* reg = erts_proc_sched_data((Proc))->registers->x_reg_array.d;	\
     (Proc)->freason = (Reason);					\
     (Proc)->current = &(Bif)->info.mfa;                         \
     reg[0] = (Eterm) (A0);					\
@@ -254,132 +310,134 @@ do {								\
     (Ret) = THE_NON_VALUE;					\
 } while (0)
 
-#define ERTS_BIF_PREP_TRAP0(Ret, Trap, Proc)	\
-do {						\
-    (Proc)->arity = 0;				\
-    (Proc)->i = (BeamInstr*) ((Trap)->addressv[erts_active_code_ix()]);	\
-    (Proc)->freason = TRAP;			\
-    (Ret) = THE_NON_VALUE;			\
-} while (0)
 
-#define ERTS_BIF_PREP_TRAP1(Ret, Trap, Proc, A0)		\
-do {								\
-    Eterm* reg = erts_proc_sched_data((Proc))->x_reg_array;	\
-    (Proc)->arity = 1;						\
-    reg[0] = (Eterm) (A0);					\
-    (Proc)->i = (BeamInstr*) ((Trap)->addressv[erts_active_code_ix()]); \
-    (Proc)->freason = TRAP;					\
-    (Ret) = THE_NON_VALUE;					\
-} while (0)
+#ifdef BEAMASM
 
-#define ERTS_BIF_PREP_TRAP2(Ret, Trap, Proc, A0, A1)		\
-do {								\
-    Eterm* reg = erts_proc_sched_data((Proc))->x_reg_array;	\
-    (Proc)->arity = 2;						\
-    reg[0] = (Eterm) (A0);					\
-    reg[1] = (Eterm) (A1);					\
-    (Proc)->i = (BeamInstr*) ((Trap)->addressv[erts_active_code_ix()]); \
-    (Proc)->freason = TRAP;					\
-    (Ret) = THE_NON_VALUE;					\
-} while (0)
+/* See `emit_bif_export_trap` for details. */
+extern ErtsCodePtr beam_bif_export_trap;
+#define ERTS_BIF_PREP_TRAP(Export, Proc, Arity)                               \
+    do {                                                                      \
+        (Proc)->i = beam_bif_export_trap;                                     \
+        (Proc)->arity = (Arity);                                              \
+        (Proc)->freason = TRAP;                                               \
+        (Proc)->current = &(Export)->info.mfa;                                \
+    } while(0);
 
-#define ERTS_BIF_PREP_TRAP3(Ret, Trap, Proc, A0, A1, A2)	\
-do {								\
-    Eterm* reg = erts_proc_sched_data((Proc))->x_reg_array;	\
-    (Proc)->arity = 3;						\
-    reg[0] = (Eterm) (A0);					\
-    reg[1] = (Eterm) (A1);					\
-    reg[2] = (Eterm) (A2);					\
-    (Proc)->i = (BeamInstr*) ((Trap)->addressv[erts_active_code_ix()]); \
-    (Proc)->freason = TRAP;					\
-    (Ret) = THE_NON_VALUE;					\
-} while (0)
+#else /* defined(BEAMASM) */
 
-#define ERTS_BIF_PREP_TRAP4(Ret, Trap, Proc, A0, A1, A2, A3)	\
-do {								\
-    Eterm* reg = erts_proc_sched_data((Proc))->x_reg_array;	\
-    (Proc)->arity = 4;						\
-    reg[0] = (Eterm) (A0);					\
-    reg[1] = (Eterm) (A1);					\
-    reg[2] = (Eterm) (A2);					\
-    reg[3] = (Eterm) (A3);					\
-    (Proc)->i = (BeamInstr*) ((Trap)->addressv[erts_active_code_ix()]); \
-    (Proc)->freason = TRAP;					\
-    (Ret) = THE_NON_VALUE;					\
-} while (0)
+#define ERTS_BIF_PREP_TRAP(Export, Proc, Arity)                               \
+    do {                                                                      \
+        (Proc)->i = (Export)->addresses[erts_active_code_ix()];               \
+        (Proc)->arity = (Arity);                                              \
+        (Proc)->freason = TRAP;                                               \
+    } while(0);
 
-#define ERTS_BIF_PREP_TRAP3_NO_RET(Trap, Proc, A0, A1, A2)\
-do {							\
-    Eterm* reg = erts_proc_sched_data((Proc))->x_reg_array;	\
-    (Proc)->arity = 3;					\
-    reg[0] = (Eterm) (A0);		\
-    reg[1] = (Eterm) (A1);		\
-    reg[2] = (Eterm) (A2);		\
-    (Proc)->i = (BeamInstr*) ((Trap)->addressv[erts_active_code_ix()]); \
-    (Proc)->freason = TRAP;				\
-} while (0)
+#endif /* !defined(BEAMASM) */
 
-#define BIF_TRAP0(Trap_, p) do {                  \
-      (p)->arity = 0;				\
-      (p)->i = (BeamInstr*) ((Trap_)->addressv[erts_active_code_ix()]);	\
-      (p)->freason = TRAP;			\
-      return THE_NON_VALUE;			\
- } while(0)
+#define ERTS_BIF_PREP_TRAP0(Ret, Trap, Proc)                                  \
+    do {                                                                      \
+        ERTS_BIF_PREP_TRAP((Trap), (Proc), 0);                                \
+        (Ret) = THE_NON_VALUE;                                                \
+    } while (0)
 
-#define BIF_TRAP1(Trap_, p, A0) do {				\
-      Eterm* reg = erts_proc_sched_data((p))->x_reg_array;	\
-      (p)->arity = 1;						\
-      reg[0] = (A0);						\
-      (p)->i = (BeamInstr*) ((Trap_)->addressv[erts_active_code_ix()]); \
-      (p)->freason = TRAP;					\
-      return THE_NON_VALUE;					\
- } while(0)
+#define ERTS_BIF_PREP_TRAP1(Ret, Trap, Proc, A0)                              \
+    do {                                                                      \
+        Eterm* reg = erts_proc_sched_data((Proc))->registers->x_reg_array.d;  \
+        ERTS_BIF_PREP_TRAP((Trap), (Proc), 1);                                \
+        reg[0] = (Eterm) (A0);                                                \
+        (Ret) = THE_NON_VALUE;                                                \
+    } while (0)
 
-#define BIF_TRAP2(Trap_, p, A0, A1) do {			\
-      Eterm* reg = erts_proc_sched_data((p))->x_reg_array;	\
-      (p)->arity = 2;						\
-      reg[0] = (A0);						\
-      reg[1] = (A1);						\
-      (p)->i = (BeamInstr*) ((Trap_)->addressv[erts_active_code_ix()]); \
-      (p)->freason = TRAP;					\
-      return THE_NON_VALUE;					\
- } while(0)
+#define ERTS_BIF_PREP_TRAP2(Ret, Trap, Proc, A0, A1)                          \
+    do {                                                                      \
+        Eterm* reg = erts_proc_sched_data((Proc))->registers->x_reg_array.d;  \
+        ERTS_BIF_PREP_TRAP((Trap), (Proc), 2);                                \
+        reg[0] = (Eterm) (A0);                                                \
+        reg[1] = (Eterm) (A1);                                                \
+        (Ret) = THE_NON_VALUE;                                                \
+    } while (0)
 
-#define BIF_TRAP3(Trap_, p, A0, A1, A2) do {			\
-      Eterm* reg = erts_proc_sched_data((p))->x_reg_array;	\
-      (p)->arity = 3;						\
-      reg[0] = (A0);						\
-      reg[1] = (A1);						\
-      reg[2] = (A2);						\
-      (p)->i = (BeamInstr*) ((Trap_)->addressv[erts_active_code_ix()]); \
-      (p)->freason = TRAP;					\
-      return THE_NON_VALUE;					\
- } while(0)
+#define ERTS_BIF_PREP_TRAP3(Ret, Trap, Proc, A0, A1, A2)                      \
+    do {                                                                      \
+        Eterm* reg = erts_proc_sched_data((Proc))->registers->x_reg_array.d;  \
+        ERTS_BIF_PREP_TRAP((Trap), (Proc), 3);                                \
+        reg[0] = (Eterm) (A0);                                                \
+        reg[1] = (Eterm) (A1);                                                \
+        reg[2] = (Eterm) (A2);                                                \
+        (Ret) = THE_NON_VALUE;                                                \
+    } while (0)
 
-#define BIF_TRAP4(Trap_, p, A0, A1, A2, A3) do {		\
-      Eterm* reg = erts_proc_sched_data((p))->x_reg_array;	\
-      (p)->arity = 4;						\
-      reg[0] = (A0);						\
-      reg[1] = (A1);						\
-      reg[2] = (A2);						\
-      reg[3] = (A3);						\
-      (p)->i = (BeamInstr*) ((Trap_)->addressv[erts_active_code_ix()]); \
-      (p)->freason = TRAP;					\
-      return THE_NON_VALUE;					\
- } while(0)
+#define ERTS_BIF_PREP_TRAP4(Ret, Trap, Proc, A0, A1, A2, A3)                  \
+    do {                                                                      \
+        Eterm* reg = erts_proc_sched_data((Proc))->registers->x_reg_array.d;  \
+        ERTS_BIF_PREP_TRAP((Trap), (Proc), 4);                                \
+        reg[0] = (Eterm) (A0);                                                \
+        reg[1] = (Eterm) (A1);                                                \
+        reg[2] = (Eterm) (A2);                                                \
+        reg[3] = (Eterm) (A3);                                                \
+        (Ret) = THE_NON_VALUE;                                                \
+    } while (0)
 
-#define BIF_TRAP_CODE_PTR_0(p, Code_) do {	\
-      (p)->arity = 0;				\
-      (p)->i = (BeamInstr*) (Code_);		\
-      (p)->freason = TRAP;			\
-      return THE_NON_VALUE;			\
- } while(0)
+#define ERTS_BIF_PREP_TRAP3_NO_RET(Trap, Proc, A0, A1, A2)                    \
+    do {                                                                      \
+        Eterm* reg = erts_proc_sched_data((Proc))->registers->x_reg_array.d;  \
+        ERTS_BIF_PREP_TRAP((Trap), (Proc), 3);                                \
+        reg[0] = (Eterm) (A0);                                                \
+        reg[1] = (Eterm) (A1);                                                \
+        reg[2] = (Eterm) (A2);                                                \
+    } while (0)
 
-#define BIF_TRAP_CODE_PTR_(p, Code_) do {	\
-      (p)-> i = (BeamInstr*) (Code_);		\
-      (p)->freason = TRAP;			\
-      return THE_NON_VALUE;			\
- } while(0)
+#define BIF_TRAP0(Trap_, p)                                                   \
+    do {                                                                      \
+        ERTS_BIF_PREP_TRAP((Trap_), (p), 0);                                  \
+        return THE_NON_VALUE;                                                 \
+    } while(0)
+
+#define BIF_TRAP1(Trap_, p, A0)                                               \
+    do {                                                                      \
+        Eterm* reg = erts_proc_sched_data((p))->registers->x_reg_array.d;     \
+        ERTS_BIF_PREP_TRAP((Trap_), (p), 1);                                  \
+        reg[0] = (A0);                                                        \
+        return THE_NON_VALUE;                                                 \
+    } while(0)
+
+#define BIF_TRAP2(Trap_, p, A0, A1)                                           \
+    do {                                                                      \
+        Eterm* reg = erts_proc_sched_data((p))->registers->x_reg_array.d;     \
+        ERTS_BIF_PREP_TRAP((Trap_), (p), 2);                                  \
+        reg[0] = (A0);                                                        \
+        reg[1] = (A1);                                                        \
+        return THE_NON_VALUE;                                                 \
+    } while(0)
+
+#define BIF_TRAP3(Trap_, p, A0, A1, A2)                                       \
+        do {                                                                  \
+        Eterm* reg = erts_proc_sched_data((p))->registers->x_reg_array.d;     \
+        ERTS_BIF_PREP_TRAP((Trap_), (p), 3);                                  \
+        reg[0] = (A0);                                                        \
+        reg[1] = (A1);                                                        \
+        reg[2] = (A2);                                                        \
+        return THE_NON_VALUE;                                                 \
+    } while(0)
+
+#define BIF_TRAP4(Trap_, p, A0, A1, A2, A3)                                   \
+    do {                                                                      \
+        Eterm* reg = erts_proc_sched_data((p))->registers->x_reg_array.d;     \
+        ERTS_BIF_PREP_TRAP((Trap_), (p), 4);                                  \
+        reg[0] = (A0);                                                        \
+        reg[1] = (A1);                                                        \
+        reg[2] = (A2);                                                        \
+        reg[3] = (A3);                                                        \
+        return THE_NON_VALUE;                                                 \
+    } while(0)
+
+#define BIF_TRAP_CODE_PTR(p, Code_, Arity_)                                   \
+    do {                                                                      \
+        (p)->arity = (Arity_);                                                \
+        (p)->i = (Code_);                                                     \
+        (p)->freason = TRAP;                                                  \
+        return THE_NON_VALUE;                                                 \
+    } while(0)
 
 extern Export bif_return_trap_export;
 #define ERTS_BIF_PREP_YIELD_RETURN_X(RET, P, VAL, OP)			\
@@ -462,6 +520,15 @@ do {									\
     BIF_TRAP4((TRP), (P), (A0), (A1), (A2), (A3));                      \
 } while (0)
 
+extern Export erts_bif_handle_signals_return_export;
+
+#define ERTS_BIF_HANDLE_SIGNALS_RETURN(P, VAL)                          \
+    BIF_TRAP1(&erts_bif_handle_signals_return_export, (P), (VAL))
+
+#define ERTS_BIF_PREP_HANDLE_SIGNALS_RETURN(Ret, P, Val)                \
+    ERTS_BIF_PREP_TRAP1((Ret), &erts_bif_handle_signals_return_export,  \
+                        (P), (Val))
+
 #define ERTS_BIF_PREP_EXITED(RET, PROC)	                                \
 do {                                                                    \
     KILL_CATCHES((PROC));                                               \
@@ -481,12 +548,13 @@ do {					\
 } while (0)
 
 int erts_call_dirty_bif(ErtsSchedulerData *esdp, Process *c_p,
-			BeamInstr *I, Eterm *reg);
+                        ErtsCodePtr I, Eterm *reg);
 
 BIF_RETTYPE
 erts_schedule_bif(Process *proc,
 		  Eterm *argv,
-		  BeamInstr *i,
+		  ErtsCodePtr i,
+		  const ErtsCodeMFA *mfa,
 		  ErtsBifFunc dbf,
 		  ErtsSchedType sched_type,
 		  Eterm mod,
@@ -496,7 +564,8 @@ erts_schedule_bif(Process *proc,
 ERTS_GLB_INLINE BIF_RETTYPE
 erts_reschedule_bif(Process *proc,
 		    Eterm *argv,
-		    BeamInstr *i,
+		    ErtsCodePtr i,
+		    const ErtsCodeMFA *mfa,
 		    ErtsBifFunc dbf,
 		    ErtsSchedType sched_type);
 
@@ -505,63 +574,18 @@ erts_reschedule_bif(Process *proc,
 ERTS_GLB_INLINE BIF_RETTYPE
 erts_reschedule_bif(Process *proc,
 		    Eterm *argv,
-		    BeamInstr *i,
+		    ErtsCodePtr i,
+		    const ErtsCodeMFA *mfa,
 		    ErtsBifFunc dbf,
 		    ErtsSchedType sched_type)
 {
-    return erts_schedule_bif(proc, argv, i, dbf, sched_type,
+    return erts_schedule_bif(proc, argv, i, mfa, dbf, sched_type,
 			     THE_NON_VALUE, THE_NON_VALUE, -1);
 }
 
 #endif /* ERTS_GLB_INLINE_INCL_FUNC_DEF */
 
-#ifdef ERL_WANT_HIPE_BIF_WRAPPER__
-
-#ifndef HIPE
-
-#define HIPE_WRAPPER_BIF_DISABLE_GC(BIF_NAME, ARITY)
-
-#else
-
-#include "erl_fun.h"
-#include "hipe_mode_switch.h"
-
-/*
- * Hipe wrappers used by native code for BIFs that disable GC while trapping.
- * Also add usage of the wrapper in ../hipe/hipe_bif_list.m4
- *
- * Problem:
- * When native code calls a BIF that traps, hipe_mode_switch will push a
- * "trap frame" on the Erlang stack in order to find its way back from beam_emu
- * back to native caller when finally done. If GC is disabled and stack/heap
- * is full there is no place to push the "trap frame".
- *
- * Solution:
- * We reserve space on stack for the "trap frame" here before the BIF is called.
- * If the BIF does not trap, the space is reclaimed here before returning.
- * If the BIF traps, hipe_push_beam_trap_frame() will detect that a "trap frame"
- * already is reserved and use it.
- */
-
-
-#define HIPE_WRAPPER_BIF_DISABLE_GC(BIF_NAME, ARITY)			\
-BIF_RETTYPE								\
-nbif_impl_hipe_wrapper_ ## BIF_NAME ## _ ## ARITY (NBIF_ALIST);		\
-BIF_RETTYPE								\
-nbif_impl_hipe_wrapper_ ## BIF_NAME ## _ ## ARITY (NBIF_ALIST)		\
-{									\
-    BIF_RETTYPE  res;							\
-    hipe_reserve_beam_trap_frame(BIF_P, BIF__ARGS, ARITY);		\
-    res = nbif_impl_ ## BIF_NAME ## _ ## ARITY (NBIF_CALL_ARGS);	\
-    if (is_value(res) || BIF_P->freason != TRAP) {			\
-	hipe_unreserve_beam_trap_frame(BIF_P);				\
-    }									\
-    return res;								\
-}
-
-#endif
-
-#endif /* ERL_WANT_HIPE_BIF_WRAPPER__ */
+Uint16 erts_monitor_opts(Eterm opts, Eterm *tag);
 
 #include "erl_bif_table.h"
 
