@@ -1,7 +1,7 @@
 /*
  * %CopyrightBegin%
  *
- * Copyright Ericsson AB 1999-2018. All Rights Reserved.
+ * Copyright Ericsson AB 1999-2021. All Rights Reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -45,8 +45,6 @@
 #define DECLARE_TMP(VariableName,N,P)  Eterm VariableName[2]
 #define ARG_IS_NOT_TMP(Arg,Tmp) ((Arg) != make_big((Tmp)))
 
-static Eterm shift(Process* p, Eterm arg1, Eterm arg2, int right);
-
 static ERTS_INLINE void maybe_shrink(Process* p, Eterm* hp, Eterm res, Uint alloc)
 {
     Uint actual;
@@ -81,7 +79,7 @@ BIF_RETTYPE splus_2(BIF_ALIST_2)
 
 BIF_RETTYPE sminus_1(BIF_ALIST_1)
 {
-    BIF_RET(erts_mixed_minus(BIF_P, make_small(0), BIF_ARG_1));
+    BIF_RET(erts_unary_minus(BIF_P, BIF_ARG_1));
 } 
 
 BIF_RETTYPE sminus_2(BIF_ALIST_2)
@@ -150,18 +148,9 @@ BIF_RETTYPE bxor_2(BIF_ALIST_2)
     BIF_RET(erts_bxor(BIF_P, BIF_ARG_1, BIF_ARG_2));
 } 
 
-BIF_RETTYPE bsl_2(BIF_ALIST_2)
-{
-    BIF_RET(shift(BIF_P, BIF_ARG_1, BIF_ARG_2, 0));
-} 
-
-BIF_RETTYPE bsr_2(BIF_ALIST_2)
-{
-    BIF_RET(shift(BIF_P, BIF_ARG_1, BIF_ARG_2, 1));
-} 
 
 static Eterm
-shift(Process* p, Eterm arg1, Eterm arg2, int right)
+erts_shift(Process* p, Eterm arg1, Eterm arg2, int right)
 {
     Sint i;
     Sint ires;
@@ -282,6 +271,24 @@ shift(Process* p, Eterm arg1, Eterm arg2, int right)
     BIF_ERROR(p, BADARITH);
 }
 
+Eterm erts_bsl(Process* p, Eterm arg1, Eterm arg2) {
+    return erts_shift(p, arg1, arg2, 0);
+}
+
+Eterm erts_bsr(Process* p, Eterm arg1, Eterm arg2) {
+    return erts_shift(p, arg1, arg2, 1);
+}
+
+BIF_RETTYPE bsl_2(BIF_ALIST_2)
+{
+    BIF_RET(erts_bsl(BIF_P, BIF_ARG_1, BIF_ARG_2));
+}
+
+BIF_RETTYPE bsr_2(BIF_ALIST_2)
+{
+    BIF_RET(erts_bsr(BIF_P, BIF_ARG_1, BIF_ARG_2));
+}
+
 BIF_RETTYPE bnot_1(BIF_ALIST_1)
 {
     Eterm ret;
@@ -454,6 +461,71 @@ erts_mixed_plus(Process* p, Eterm arg1, Eterm arg2)
     }
 }
 
+/*
+ * While "-value" is generally the same as "0 - value",
+ * that's not true for floats due to positive and negative
+ * zeros, so we implement unary minus as its own operation.
+ */
+Eterm
+erts_unary_minus(Process* p, Eterm arg)
+{
+    Eterm hdr, res;
+    FloatDef f;
+    dsize_t sz;
+    int need_heap;
+    Eterm* hp;
+    Sint ires;
+
+    ERTS_FP_CHECK_INIT(p);
+    switch (arg & _TAG_PRIMARY_MASK) {
+    case TAG_PRIMARY_IMMED1:
+        switch ((arg & _TAG_IMMED1_MASK) >> _TAG_PRIMARY_SIZE) {
+        case (_TAG_IMMED1_SMALL >> _TAG_PRIMARY_SIZE):
+            ires = -signed_val(arg);
+            if (IS_SSMALL(ires)) {
+                return make_small(ires);
+            } else {
+                hp = HeapFragOnlyAlloc(p, 2);
+                res = small_to_big(ires, hp);
+                return res;
+            }
+        default:
+        badarith:
+            p->freason = BADARITH;
+            return THE_NON_VALUE;
+        }
+    case TAG_PRIMARY_BOXED:
+        hdr = *boxed_val(arg);
+        switch ((hdr & _TAG_HEADER_MASK) >> _TAG_PRIMARY_SIZE) {
+        case (_TAG_HEADER_POS_BIG >> _TAG_PRIMARY_SIZE):
+        case (_TAG_HEADER_NEG_BIG >> _TAG_PRIMARY_SIZE): {
+            Eterm zero_buf[2] = {make_pos_bignum_header(1), 0};
+            Eterm zero = make_big(zero_buf);
+            sz = big_size(arg);
+            need_heap = BIG_NEED_SIZE(sz);
+            hp = HeapFragOnlyAlloc(p, need_heap);
+            res = big_minus(zero, arg, hp);
+            maybe_shrink(p, hp, res, need_heap);
+            ASSERT(is_not_nil(res));
+            return res;
+        }
+        case (_TAG_HEADER_FLOAT >> _TAG_PRIMARY_SIZE):
+            GET_DOUBLE(arg, f);
+            f.fd = -f.fd;
+            ERTS_FP_ERROR(p, f.fd, goto badarith);
+            hp = HeapFragOnlyAlloc(p, FLOAT_SIZE_OBJECT);
+            res = make_float(hp);
+            PUT_DOUBLE(f, hp);
+            return res;
+        default:
+            goto badarith;
+        }
+    default:
+        goto badarith;
+    }
+}
+
+
 Eterm
 erts_mixed_minus(Process* p, Eterm arg1, Eterm arg2)
 {
@@ -912,6 +984,121 @@ erts_mixed_div(Process* p, Eterm arg1, Eterm arg2)
     }
 }
 
+static void div_rem_shrink(Process *p,
+                           Eterm *left_hp, Eterm left,
+                           Eterm *right_hp, Eterm right)
+{
+    Uint left_size, right_size;
+
+    ASSERT(left_hp < right_hp);
+    ASSERT(!(p->heap <= left_hp && left_hp < p->htop));
+    ASSERT(!(p->heap <= right_hp && right_hp < p->htop));
+
+    left_size = size_object(left);
+    right_size = size_object(right);
+
+    if (right_size == 0) {
+        /* The right term's an immediate, so we can shrink the fragment down to
+         * the left term. */
+        erts_heap_frag_shrink(p, &left_hp[left_size]);
+    } else {
+        /* The right term's a bignum, so we can't shave more than a few words
+         * off the right regardless of how small the left term is. We also need
+         * to fill the surplus words in the left with a dummy bignum to prevent
+         * holes. */
+        erts_heap_frag_shrink(p, &right_hp[right_size]);
+
+        if (&left_hp[left_size] < right_hp) {
+            Uint unused = right_hp - &left_hp[left_size] - 1;
+            left_hp[left_size] = make_pos_bignum_header(unused);
+        }
+    }
+}
+
+int erts_int_div_rem(Process* p, Eterm arg1, Eterm arg2, Eterm *q, Eterm *r)
+{
+    Eterm quotient, remainder;
+    Eterm lhs, rhs;
+    int cmp;
+
+    DECLARE_TMP(tmp_big1,0,p);
+    DECLARE_TMP(tmp_big2,1,p);
+
+    lhs = arg1;
+    rhs = arg2;
+
+    switch (NUMBER_CODE(lhs, rhs)) {
+    case SMALL_SMALL:
+        /* This case occurs if the most negative fixnum is divided by -1. */
+        ASSERT(rhs == make_small(-1));
+        lhs = small_to_big(signed_val(lhs), tmp_big1);
+
+        /* ! Fall through ! */
+    case BIG_SMALL:
+        rhs = small_to_big(signed_val(rhs), tmp_big2);
+        break;
+    case SMALL_BIG:
+        if (lhs != make_small(MIN_SMALL)) {
+            *q = SMALL_ZERO;
+            *r = lhs;
+            return 1;
+        }
+
+        lhs = small_to_big(signed_val(lhs), tmp_big1);
+        break;
+    case BIG_BIG:
+        break;
+    default:
+        p->freason = BADARITH;
+        return 0;
+    }
+
+    cmp = big_ucomp(lhs, rhs);
+
+    if (cmp < 0) {
+        quotient = SMALL_ZERO;
+        remainder = arg1;
+    } else if (cmp == 0) {
+        quotient = (big_sign(lhs) == big_sign(rhs)) ?
+                    SMALL_ONE : SMALL_MINUS_ONE;
+        remainder = SMALL_ZERO;
+    } else {
+        int lhs_size, rhs_size;
+        Uint q_need, r_need;
+        Eterm *q_hp, *r_hp;
+
+        lhs_size = big_size(lhs);
+        rhs_size = big_size(rhs);
+
+        q_need = BIG_NEED_SIZE(lhs_size - rhs_size + 1);
+        r_need = BIG_NEED_SIZE(lhs_size);
+
+        q_hp = HeapFragOnlyAlloc(p, q_need + r_need);
+        r_hp = q_hp + q_need;
+
+        if (!big_div_rem(lhs, rhs, q_hp, &quotient, r_hp, &remainder)) {
+            ASSERT(is_non_value(erts_int_div(p, arg1, arg2)));
+            ASSERT(is_non_value(erts_int_rem(p, arg1, arg2)));
+
+            erts_heap_frag_shrink(p, q_hp);
+            p->freason = SYSTEM_LIMIT;
+            return 0;
+        }
+
+        ASSERT(q_need + r_need >= size_object(quotient) + size_object(remainder));
+
+        div_rem_shrink(p, q_hp, quotient, r_hp, remainder);
+    }
+
+    ASSERT(eq(erts_int_div(p, arg1, arg2), quotient));
+    ASSERT(eq(erts_int_rem(p, arg1, arg2), remainder));
+
+    *q = quotient;
+    *r = remainder;
+
+    return 1;
+}
+
 Eterm
 erts_int_div(Process* p, Eterm arg1, Eterm arg2)
 {
@@ -972,6 +1159,11 @@ erts_int_rem(Process* p, Eterm arg1, Eterm arg2)
     int ires;
 
     switch (NUMBER_CODE(arg1, arg2)) {
+    case SMALL_SMALL:
+	/* This case occurs if the most negative fixnum is divided by -1. */
+	ASSERT(arg2 == make_small(-1));
+	arg1 = small_to_big(signed_val(arg1), tmp_big1);
+	/*FALLTHROUGH*/
     case BIG_SMALL:
 	arg2 = small_to_big(signed_val(arg2), tmp_big2);
 	goto L_big_rem;
@@ -1108,11 +1300,11 @@ Eterm erts_bnot(Process* p, Eterm arg)
 	maybe_shrink(p, bigp, ret, need);
 	if (is_nil(ret)) {
 	    p->freason = SYSTEM_LIMIT;
-	    return NIL;
+	    return THE_NON_VALUE;
 	}
     } else {
 	p->freason = BADARITH;
-	return NIL;
+	return THE_NON_VALUE;
     }
     return ret;
 } 
