@@ -1,7 +1,7 @@
 /*
  * %CopyrightBegin%
  *
- * Copyright Ericsson AB 2008-2018. All Rights Reserved.
+ * Copyright Ericsson AB 2008-2021. All Rights Reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -327,7 +327,7 @@ crc32_1(BIF_ALIST_1)
     res_sum = erts_make_integer(chksum,BIF_P);
     if (rest != NIL) {
 	BUMP_ALL_REDS(BIF_P);
-	BIF_TRAP2(&bif_trap_export[BIF_crc32_2], BIF_P, res_sum, rest);
+	BIF_TRAP2(BIF_TRAP_EXPORT(BIF_crc32_2), BIF_P, res_sum, rest);
     }
     BIF_RET(res_sum);
 }
@@ -354,7 +354,7 @@ crc32_2(BIF_ALIST_2)
     res_sum = erts_make_integer(chksum,BIF_P);
     if (rest != NIL) {
 	BUMP_ALL_REDS(BIF_P);
-	BIF_TRAP2(&bif_trap_export[BIF_crc32_2], BIF_P, res_sum, rest);
+	BIF_TRAP2(BIF_TRAP_EXPORT(BIF_crc32_2), BIF_P, res_sum, rest);
     }
     BIF_RET(res_sum);
 }
@@ -407,7 +407,7 @@ adler32_1(BIF_ALIST_1)
     res_sum = erts_make_integer(chksum,BIF_P);
     if (rest != NIL) {
 	BUMP_ALL_REDS(BIF_P);
-	BIF_TRAP2(&bif_trap_export[BIF_adler32_2], BIF_P, res_sum, rest);
+	BIF_TRAP2(BIF_TRAP_EXPORT(BIF_adler32_2), BIF_P, res_sum, rest);
     }
     BIF_RET(res_sum);
 }
@@ -434,7 +434,7 @@ adler32_2(BIF_ALIST_2)
     res_sum = erts_make_integer(chksum,BIF_P);
     if (rest != NIL) {
 	BUMP_ALL_REDS(BIF_P);
-	BIF_TRAP2(&bif_trap_export[BIF_adler32_2], BIF_P, res_sum, rest);
+	BIF_TRAP2(BIF_TRAP_EXPORT(BIF_adler32_2), BIF_P, res_sum, rest);
     }
     BIF_RET(res_sum);
 }
@@ -575,7 +575,7 @@ md5_update_2(BIF_ALIST_2)
     bin = new_binary(BIF_P, (byte *) &context, sizeof(MD5_CTX));
     if (rest != NIL) {
 	BUMP_ALL_REDS(BIF_P);
-	BIF_TRAP2(&bif_trap_export[BIF_md5_update_2], BIF_P, bin, rest);
+	BIF_TRAP2(BIF_TRAP_EXPORT(BIF_md5_update_2), BIF_P, bin, rest);
     }
     BUMP_REDS(BIF_P,res);
     BIF_RET(bin);
