@@ -1,7 +1,7 @@
 /*
  * %CopyrightBegin%
  * 
- * Copyright Ericsson AB 2006-2018. All Rights Reserved.
+ * Copyright Ericsson AB 2006-2021. All Rights Reserved.
  * 
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -1554,11 +1554,9 @@ static int do_unload_driver_entry(DE_Handle *dh, Eterm *save_name)
 	    /* Future locking problems? Don't dare to let go of the
 	       diver_list lock here!*/
 	    if (q->finish) {
-		int fpe_was_unmasked = erts_block_fpe();
 		DTRACE1(driver_finish, q->name);
                 LTTNG1(driver_finish, q->name);
 		(*(q->finish))();
-		erts_unblock_fpe(fpe_was_unmasked);
 	    }
 	    erts_sys_ddll_close(dh->handle);
 	    erts_destroy_driver(q);
