@@ -1,7 +1,7 @@
 /*
  * %CopyrightBegin%
  * 
- * Copyright Ericsson AB 2018. All Rights Reserved.
+ * Copyright Ericsson AB 2018-2021. All Rights Reserved.
  * 
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -396,6 +396,12 @@
 #include "erl_proc_sig_queue.h"
 #undef ERTS_PROC_SIG_QUEUE_TYPE_ONLY
 
+#define ERL_THR_PROGRESS_TSD_TYPE_ONLY
+#include "erl_thr_progress.h"
+#undef ERL_THR_PROGRESS_TSD_TYPE_ONLY
+
+#include "erl_alloc.h"
+
 #if defined(DEBUG) || 0
 #  define ERTS_ML_DEBUG
 #else
@@ -408,7 +414,18 @@
 #  define ERTS_ML_ASSERT(E) ((void) 1)
 #endif
 
-#define ERTS_MON_TYPE_MAX               ((Uint16) 7)
+#define ERTS_ML_STATE_ALIAS_BITS        2
+#define ERTS_ML_STATE_ALIAS_SHIFT       11
+#define ERTS_ML_STATE_ALIAS_MASK        \
+    ((((Uint16) 1 << ERTS_ML_STATE_ALIAS_BITS) - 1) \
+     << ERTS_ML_STATE_ALIAS_SHIFT)
+
+#define ERTS_ML_STATE_ALIAS_NONE        (((Uint16) 0) << ERTS_ML_STATE_ALIAS_SHIFT)
+#define ERTS_ML_STATE_ALIAS_UNALIAS     (((Uint16) 1) << ERTS_ML_STATE_ALIAS_SHIFT)
+#define ERTS_ML_STATE_ALIAS_DEMONITOR   (((Uint16) 2) << ERTS_ML_STATE_ALIAS_SHIFT)
+#define ERTS_ML_STATE_ALIAS_ONCE        (((Uint16) 3) << ERTS_ML_STATE_ALIAS_SHIFT)
+
+#define ERTS_MON_TYPE_MAX               ((Uint16) 8)
 
 #define ERTS_MON_TYPE_PROC              ((Uint16) 0)
 #define ERTS_MON_TYPE_PORT              ((Uint16) 1)
@@ -417,7 +434,8 @@
 #define ERTS_MON_TYPE_RESOURCE          ((Uint16) 4)
 #define ERTS_MON_TYPE_NODE              ((Uint16) 5)
 #define ERTS_MON_TYPE_NODES             ((Uint16) 6)
-#define ERTS_MON_TYPE_SUSPEND           ERTS_MON_TYPE_MAX
+#define ERTS_MON_TYPE_SUSPEND           ((Uint16) 7)
+#define ERTS_MON_TYPE_ALIAS             ERTS_MON_TYPE_MAX
 
 #define ERTS_MON_LNK_TYPE_MAX           (ERTS_MON_TYPE_MAX + ((Uint16) 3))
 #define ERTS_LNK_TYPE_MAX               ERTS_MON_LNK_TYPE_MAX
@@ -431,9 +449,25 @@
 #define ERTS_ML_FLG_IN_SUBTABLE         (((Uint16) 1) << 2)
 #define ERTS_ML_FLG_NAME                (((Uint16) 1) << 3)
 #define ERTS_ML_FLG_EXTENDED            (((Uint16) 1) << 4)
+#define ERTS_ML_FLG_SPAWN_PENDING       (((Uint16) 1) << 5)
+#define ERTS_ML_FLG_SPAWN_MONITOR       (((Uint16) 1) << 6)
+#define ERTS_ML_FLG_SPAWN_LINK          (((Uint16) 1) << 7)
+#define ERTS_ML_FLG_SPAWN_ABANDONED     (((Uint16) 1) << 8)
+#define ERTS_ML_FLG_SPAWN_NO_SMSG       (((Uint16) 1) << 9)
+#define ERTS_ML_FLG_SPAWN_NO_EMSG       (((Uint16) 1) << 10)
+#define ERTS_ML_FLG_ALIAS_BIT1          (((Uint16) 1) << 11)
+#define ERTS_ML_FLG_ALIAS_BIT2          (((Uint16) 1) << 12)
+#define ERTS_ML_FLG_TAG                 (((Uint16) 1) << 13)
 
 #define ERTS_ML_FLG_DBG_VISITED         (((Uint16) 1) << 15)
 
+#define ERTS_ML_FLGS_SPAWN              (ERTS_ML_FLG_SPAWN_PENDING      \
+                                         | ERTS_ML_FLG_SPAWN_MONITOR    \
+                                         | ERTS_ML_FLG_SPAWN_LINK       \
+                                         | ERTS_ML_FLG_SPAWN_ABANDONED  \
+                                         | ERTS_ML_FLG_SPAWN_NO_SMSG    \
+                                         | ERTS_ML_FLG_SPAWN_NO_EMSG)
+
 /* Flags that should be the same on both monitor/link halves */
 #define ERTS_ML_FLGS_SAME \
     (ERTS_ML_FLG_EXTENDED|ERTS_ML_FLG_NAME)
@@ -468,7 +502,7 @@ struct ErtsMonLnkNode__ {
     Uint16 type;
 };
 
-typedef struct {
+typedef struct ErtsMonLnkDist__ {
     Eterm nodename;
     Uint32 connection_id;
     erts_atomic_t refc;
@@ -478,6 +512,8 @@ typedef struct {
     ErtsMonLnkNode *monitors; /* Monitor double linked circular list */
     ErtsMonLnkNode *orig_name_monitors; /* Origin named monitors
                                            read-black tree */
+    ErtsMonLnkNode *dist_pend_spawn_exit;
+    ErtsThrPrgrLaterOp cleanup_lop;
 } ErtsMonLnkDist;
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
@@ -530,7 +566,7 @@ ERTS_GLB_INLINE void erts_ml_dl_list_delete__(ErtsMonLnkNode **list,
                                               ErtsMonLnkNode *ml);
 ERTS_GLB_INLINE ErtsMonLnkNode *erts_ml_dl_list_first__(ErtsMonLnkNode *list);
 ERTS_GLB_INLINE ErtsMonLnkNode *erts_ml_dl_list_last__(ErtsMonLnkNode *list);
-void erts_mon_link_dist_destroy__(ErtsMonLnkDist *mld);
+void erts_schedule_mon_link_dist_destruction__(ErtsMonLnkDist *mld);
 ERTS_GLB_INLINE void *erts_ml_node_to_main_struct__(ErtsMonLnkNode *mln);
 
 /* implementations for globally inlined misc functions... */
@@ -548,7 +584,7 @@ erts_mon_link_dist_dec_refc(ErtsMonLnkDist *mld)
 {
     ERTS_ML_ASSERT(erts_atomic_read_nob(&mld->refc) > 0);
     if (erts_atomic_dec_read_nob(&mld->refc) == 0)
-        erts_mon_link_dist_destroy__(mld);
+        erts_schedule_mon_link_dist_destruction__(mld);
 }
 
 ERTS_GLB_INLINE void *
@@ -627,16 +663,24 @@ typedef int (*ErtsMonitorFunc)(ErtsMonitor *, void *, Sint);
 
 typedef struct {
     ErtsMonitor origin;
-    ErtsMonitor target;
+    union {
+        ErtsMonitor target;
+        Eterm ref_heap[ERTS_MAX_INTERNAL_REF_SIZE];
+    } u;
     Eterm ref;
     erts_atomic32_t refc;
 } ErtsMonitorData;
 
 typedef struct {
     ErtsMonitorData md;
-    ErtsORefThing oref_thing;
+    Eterm ref_heap[ERTS_MAX_INTERNAL_REF_SIZE];
 } ErtsMonitorDataHeap;
 
+typedef struct {
+    ErtsMonitorData md;
+    Eterm heap[1 + ERTS_MAX_INTERNAL_REF_SIZE];
+} ErtsMonitorDataTagHeap;
+
 typedef struct ErtsMonitorDataExtended__ ErtsMonitorDataExtended;
 
 struct ErtsMonitorDataExtended__ {
@@ -700,7 +744,7 @@ ErtsMonitor *erts_monitor_tree_lookup(ErtsMonitor *root, Eterm key);
  * @returns                     Pointer to a monitor with the
  *                              key 'key'. If no monitor with the key
  *                              'key' was found and 'mon' was inserted
- *                              'mon' is returned.
+ *                              'NULL' is returned.
  *
  */
 ErtsMonitor *erts_monotor_tree_lookup_insert(ErtsMonitor **root,
@@ -767,11 +811,11 @@ void erts_monitor_tree_insert(ErtsMonitor **root, ErtsMonitor *mon);
  *
  * @param[in]     old           Monitor to remove from the tree
  *
- * @param[in]     new           Monitor to insert into the tree
+ * @param[in]     new_           Monitor to insert into the tree
  *
  */
 void erts_monitor_tree_replace(ErtsMonitor **root, ErtsMonitor *old,
-                               ErtsMonitor *new);
+                               ErtsMonitor *new_);
 
 /**
  *
@@ -1147,11 +1191,15 @@ int erts_monitor_list_foreach_delete_yielding(ErtsMonitor **list,
  *
  * @param[in]     name          An atom (the name) or NIL depending on type
  *
+ * @param[in]     tag           Tag to use in message when monitor is
+ *                              triggered or THE_NON_VALUE if default
+ *                              should be used.
+ *
  * @returns                     A pointer to monitor data structure
  *
  */
 ErtsMonitorData *erts_monitor_create(Uint16 type, Eterm ref, Eterm origin,
-                                     Eterm target, Eterm name);
+                                     Eterm target, Eterm name, Eterm tag);
 
 /**
  *
@@ -1370,21 +1418,24 @@ extern size_t erts_monitor_node_key_offset;
 ERTS_GLB_INLINE ErtsMonitorData *
 erts_monitor_to_data(ErtsMonitor *mon)
 {
-    ErtsMonitorData *mdp = erts_ml_node_to_main_struct__((ErtsMonLnkNode *) mon);
+    ErtsMonitorData *mdp = (ErtsMonitorData *)erts_ml_node_to_main_struct__((ErtsMonLnkNode *) mon);
 
 #ifdef ERTS_ML_DEBUG
     ERTS_ML_ASSERT(!(mdp->origin.flags & ERTS_ML_FLG_TARGET));
     ERTS_ML_ASSERT(erts_monitor_origin_offset == (size_t) mdp->origin.offset);
-    ERTS_ML_ASSERT(!!(mdp->target.flags & ERTS_ML_FLG_TARGET));
-    ERTS_ML_ASSERT(erts_monitor_target_offset == (size_t) mdp->target.offset);
+    ERTS_ML_ASSERT(mon->type == ERTS_MON_TYPE_ALIAS
+                   || !!(mdp->u.target.flags & ERTS_ML_FLG_TARGET));
+    ERTS_ML_ASSERT(mon->type == ERTS_MON_TYPE_ALIAS
+                   || erts_monitor_target_offset == (size_t) mdp->u.target.offset);
     if (mon->type == ERTS_MON_TYPE_NODE || mon->type == ERTS_MON_TYPE_NODES
         || mon->type == ERTS_MON_TYPE_SUSPEND) {
         ERTS_ML_ASSERT(erts_monitor_node_key_offset == (size_t) mdp->origin.key_offset);
-        ERTS_ML_ASSERT(erts_monitor_node_key_offset == (size_t) mdp->target.key_offset);
+        ERTS_ML_ASSERT(erts_monitor_node_key_offset == (size_t) mdp->u.target.key_offset);
     }
     else {
         ERTS_ML_ASSERT(erts_monitor_origin_key_offset == (size_t) mdp->origin.key_offset);
-        ERTS_ML_ASSERT(erts_monitor_target_key_offset == (size_t) mdp->target.key_offset);
+        ERTS_ML_ASSERT(mon->type == ERTS_MON_TYPE_ALIAS
+                       || erts_monitor_target_key_offset == (size_t) mdp->u.target.key_offset);
     }
 #endif
 
@@ -1397,9 +1448,10 @@ erts_monitor_release(ErtsMonitor *mon)
     ErtsMonitorData *mdp = erts_monitor_to_data(mon);
     ERTS_ML_ASSERT(erts_atomic32_read_nob(&mdp->refc) > 0);
 
-    if (erts_atomic32_dec_read_nob(&mdp->refc) == 0) {
+    if (erts_atomic32_dec_read_mb(&mdp->refc) == 0) {
         ERTS_ML_ASSERT(!(mdp->origin.flags & ERTS_ML_FLG_IN_TABLE));
-        ERTS_ML_ASSERT(!(mdp->target.flags & ERTS_ML_FLG_IN_TABLE));
+        ERTS_ML_ASSERT(mon->type == ERTS_MON_TYPE_ALIAS
+                       || !(mdp->u.target.flags & ERTS_ML_FLG_IN_TABLE));
 
         erts_monitor_destroy__(mdp);
     }
@@ -1409,12 +1461,12 @@ ERTS_GLB_INLINE void
 erts_monitor_release_both(ErtsMonitorData *mdp)
 {
     ERTS_ML_ASSERT((mdp->origin.flags & ERTS_ML_FLGS_SAME)
-                   == (mdp->target.flags & ERTS_ML_FLGS_SAME));
+                   == (mdp->u.target.flags & ERTS_ML_FLGS_SAME));
     ERTS_ML_ASSERT(erts_atomic32_read_nob(&mdp->refc) >= 2);
 
-    if (erts_atomic32_add_read_nob(&mdp->refc, (erts_aint32_t) -2) == 0) {
+    if (erts_atomic32_add_read_mb(&mdp->refc, (erts_aint32_t) -2) == 0) {
         ERTS_ML_ASSERT(!(mdp->origin.flags & ERTS_ML_FLG_IN_TABLE));
-        ERTS_ML_ASSERT(!(mdp->target.flags & ERTS_ML_FLG_IN_TABLE));
+        ERTS_ML_ASSERT(!(mdp->u.target.flags & ERTS_ML_FLG_IN_TABLE));
 
         erts_monitor_destroy__(mdp);
     }
@@ -1434,14 +1486,14 @@ erts_monitor_dist_insert(ErtsMonitor *mon, ErtsMonLnkDist *dist)
 
     ERTS_ML_ASSERT(!mdep->dist);
     ERTS_ML_ASSERT(dist);
-    mdep->dist = dist;
-
-    erts_mon_link_dist_inc_refc(dist);
 
     erts_mtx_lock(&dist->mtx);
 
     insert = dist->alive;
     if (insert) {
+        mdep->dist = dist;
+        erts_mon_link_dist_inc_refc(dist);
+
         if ((mon->flags & (ERTS_ML_FLG_NAME
                            | ERTS_ML_FLG_TARGET)) == ERTS_ML_FLG_NAME)
             erts_monitor_tree_insert(&dist->orig_name_monitors, mon);
@@ -1460,7 +1512,7 @@ erts_monitor_dist_delete(ErtsMonitor *mon)
     ErtsMonitorDataExtended *mdep;
     ErtsMonLnkDist *dist;
     Uint16 flags;
-    int delete;
+    int delete_;
 
     ERTS_ML_ASSERT(mon->flags & ERTS_ML_FLG_EXTENDED);
     ERTS_ML_ASSERT(mon->type == ERTS_MON_TYPE_DIST_PROC
@@ -1473,8 +1525,8 @@ erts_monitor_dist_delete(ErtsMonitor *mon)
     erts_mtx_lock(&dist->mtx);
 
     flags = mon->flags;
-    delete = !!dist->alive & !!(flags & ERTS_ML_FLG_IN_TABLE);
-    if (delete) {
+    delete_ = !!dist->alive & !!(flags & ERTS_ML_FLG_IN_TABLE);
+    if (delete_) {
         if ((flags & (ERTS_ML_FLG_NAME
                       | ERTS_ML_FLG_TARGET)) == ERTS_ML_FLG_NAME)
             erts_monitor_tree_delete(&dist->orig_name_monitors, mon);
@@ -1484,7 +1536,7 @@ erts_monitor_dist_delete(ErtsMonitor *mon)
 
     erts_mtx_unlock(&dist->mtx);
 
-    return delete;
+    return delete_;
 }
 
 
@@ -1528,18 +1580,26 @@ typedef struct ErtsMonLnkNode__ ErtsLink;
 
 typedef int (*ErtsLinkFunc)(ErtsLink *, void *, Sint);
 
+/* Internal Link */
 typedef struct {
-    ErtsLink a;
-    ErtsLink b;
+    ErtsLink link;
+    Uint64 unlinking;
+} ErtsILink;
+
+typedef struct {
+    ErtsLink proc;
+    ErtsLink dist;
     erts_atomic32_t refc;
 } ErtsLinkData;
 
+/* External Link */
 typedef struct {
     ErtsLinkData ld;
     struct erl_off_heap_header *ohhp;
     ErtsMonLnkDist *dist;
+    Uint64 unlinking;
     Eterm heap[1]; /* heap start... */
-} ErtsLinkDataExtended;
+} ErtsELink;
 
 /*
  * --- Link tree operations ---
@@ -1577,14 +1637,14 @@ ErtsLink *erts_link_tree_lookup(ErtsLink *root, Eterm item);
  * @returns                     Pointer to a link with the
  *                              key 'key'. If no link with the key
  *                              'key' was found and 'lnk' was inserted
- *                              'lnk' is returned.
+ *                              'NULL' is returned.
  *
  */
 ErtsLink *erts_link_tree_lookup_insert(ErtsLink **root, ErtsLink *lnk);
 
 /**
  *
- * @brief Lookup or create a link in a link tree.
+ * @brief Lookup or create an external link in a link tree.
  *
  * Looks up a link with the key 'other' in the link tree. If it is not
  * found, creates and insert a link with the key 'other'.
@@ -1603,9 +1663,40 @@ ErtsLink *erts_link_tree_lookup_insert(ErtsLink **root, ErtsLink *lnk);
  *
  * @param[in]     other         Id of other entity
  *
+ * @returns                     Pointer to either an already existing
+ *                              link in the tree or a newly created
+ *                              and inserted link.
+ *
  */
-ErtsLink *erts_link_tree_lookup_create(ErtsLink **root, int *created,
-                                       Uint16 type, Eterm this, Eterm other);
+ErtsLink *erts_link_external_tree_lookup_create(ErtsLink **root, int *created,
+                                                Uint16 type, Eterm this_, Eterm other);
+
+/**
+ *
+ * @brief Lookup or create an internal link in a link tree.
+ *
+ * Looks up a link with the key 'other' in the link tree. If it is not
+ * found, creates and insert a link with the key 'other'.
+ *
+ * @param[in,out] root          Pointer to pointer to root of link tree
+ *
+ * @param[out]    created       Pointer to integer. The integer is set to
+ *                              a non-zero value if no link with key
+ *                              'other' was found, and a new link
+ *                              was created. If a link was found, it
+ *                              is set to zero.
+ *
+ * @param[in]     type          Type of link
+ *
+ * @param[in]     other         Id of other entity
+ *
+ * @returns                     Pointer to either an already existing
+ *                              link in the tree or a newly created
+ *                              and inserted link.
+ *
+ */
+ErtsLink *erts_link_internal_tree_lookup_create(ErtsLink **root, int *created,
+                                                Uint16 type, Eterm other);
 
 /**
  *
@@ -1640,7 +1731,7 @@ void erts_link_tree_insert(ErtsLink **root, ErtsLink *lnk);
  * @param[in]     new           Link to insert into the tree
  *
  */
-void erts_link_tree_replace(ErtsLink **root, ErtsLink *old, ErtsLink *new);
+void erts_link_tree_replace(ErtsLink **root, ErtsLink *old, ErtsLink *new_);
 
 /**
  *
@@ -2032,60 +2123,74 @@ int erts_link_list_foreach_delete_yielding(ErtsLink **list,
 
 /**
  *
- * @brief Create a link
+ * @brief Create an external link
  *
- * Can create all types of links
+ * An external link structure contains two links, one for usage in
+ * the link tree of the process and one for usage in the dist entry.
  *
- * When the function is called it is assumed that:
- * - 'ref' is an internal ordinary reference if type is ERTS_MON_TYPE_PROC,
- *   ERTS_MON_TYPE_PORT, ERTS_MON_TYPE_TIME_OFFSET, or ERTS_MON_TYPE_RESOURCE
- * - 'ref' is NIL if type is ERTS_MON_TYPE_NODE or ERTS_MON_TYPE_NODES
- * - 'ref' is and ordinary internal reference or an external reference if
- *   type is ERTS_MON_TYPE_DIST_PROC
- * - 'name' is an atom or NIL if type is ERTS_MON_TYPE_PROC,
- *   ERTS_MON_TYPE_PORT, or ERTS_MON_TYPE_DIST_PROC
- * - 'name is NIL if type is ERTS_MON_TYPE_TIME_OFFSET, ERTS_MON_TYPE_RESOURCE,
- *   ERTS_MON_TYPE_NODE, or ERTS_MON_TYPE_NODES
- * If the above is not true, bad things will happen.
  *
- * @param[in]     type          ERTS_MON_TYPE_PROC, ERTS_MON_TYPE_PORT,
- *                              ERTS_MON_TYPE_TIME_OFFSET, ERTS_MON_TYPE_DIST_PROC,
- *                              ERTS_MON_TYPE_RESOURCE, ERTS_MON_TYPE_NODE,
- *                              or ERTS_MON_TYPE_NODES
+ * @param[in]     type          ERTS_MON_TYPE_DIST_PROC
  *
- * @param[in]     a             The key of entity a. Link structure a will
- *                              have field other.item set to 'b'.
+ * @param[in]     this          The process identifier of the local
+ *                              process. The link structure in the
+ *                              'dist' field a will have its
+ *                              'other.item' field set to 'this'.
+ *                              The 'dist' link structure is to be
+ *                              inserted on the distribution entry.
  *
- * @param[in]     b             The key of entity b. Link structure b will
- *                              have field other.item set to 'a'.
+ * @param[in]     other         The process identifier of the remote
+ *                              process. The link structure in the
+ *                              'proc' field a will have its
+ *                              'other.item' field set to 'other'.
+ *                              The 'proc' link structure is to be
+ *                              inserted on the local process.
+ *
+ * @returns                     A pointer to the link data structure
+ *                              containing the link structures. The
+ *                              link data structure is in turn part
+ *                              of the external link structure
+ *                              (ErtsELink).
  *
  */
-ErtsLinkData *erts_link_create(Uint16 type, Eterm a, Eterm b);
+ErtsLinkData *erts_link_external_create(Uint16 type, Eterm this_, Eterm other);
 
 /**
  *
- * @brief Get pointer to link data structure
+ * @brief Create an internal link
+ *
+ * @param[in]     type          ERTS_MON_TYPE_PROC, ERTS_MON_TYPE_PORT,
+ *
+ * @param[in]     id            Id of the entity linked.
+ *
+ * @returns                     A pointer to the link stucture.
+ */
+ErtsLink *erts_link_internal_create(Uint16 type, Eterm id);
+
+/**
+ *
+ * @brief Get pointer to external link data structure
  *
  * @param[in]    lnk            Pointer to link
  *
- * @returns                     Pointer to link data structure
+ * @returns                     Pointer to external link structure
  *
  */
-ERTS_GLB_INLINE ErtsLinkData *erts_link_to_data(ErtsLink *lnk);
+ERTS_GLB_INLINE ErtsELink *erts_link_to_elink(ErtsLink *lnk);
 
 /**
  *
  * @brief Get pointer to the other link structure part of the link
  *
- * @param[in]    lnk            Pointer to link
+ * @param[in]    lnk            Pointer to link structure
  *
- * @param[out]   ldpp           Pointer to pointer to link data structure,
- *                              if a non-NULL value is passed in the call
+ * @param[out]   elnkpp         Pointer to pointer to external link
+ *                              data structure, if a non-NULL value
+ *                              is passed in the call
  *
- * @returns                     Pointer to other link
+ * @returns                     Pointer to other link structure
  *
  */
-ERTS_GLB_INLINE ErtsLink *erts_link_to_other(ErtsLink *lnk, ErtsLinkData **ldpp);
+ERTS_GLB_INLINE ErtsLink *erts_link_to_other(ErtsLink *lnk, ErtsELink **elnkpp);
 
 /**
  *
@@ -2099,12 +2204,29 @@ ERTS_GLB_INLINE ErtsLink *erts_link_to_other(ErtsLink *lnk, ErtsLinkData **ldpp)
  */
 ERTS_GLB_INLINE int erts_link_is_in_table(ErtsLink *lnk);
 
+/**
+ *
+ * @brief Release an internal link
+ *
+ * When the function is called it is assumed that:
+ * - 'lnk' link is not part of any list or tree
+ * - 'lnk' is not referred to by any other structures
+ * If the above are not true, bad things will happen.
+ *
+ * @param[in]    lnk            Pointer to link
+ *
+ */
+ERTS_GLB_INLINE void erts_link_internal_release(ErtsLink *lnk);
+
 /**
  *
  * @brief Release link
  *
- * When both link halves part of the link have been released the link
- * structure will be deallocated.
+ * Can be used to release a link half of an external
+ * link as well as an internal link. In the external
+ * case both link halves part of the external link have
+ * to been released before the link structure will be
+ * deallocated.
  *
  * When the function is called it is assumed that:
  * - 'lnk' link is not part of any list or tree
@@ -2118,10 +2240,11 @@ ERTS_GLB_INLINE void erts_link_release(ErtsLink *lnk);
 
 /**
  *
- * @brief Release both link halves of a link simultaneously
+ * @brief Release both link halves of an external link
+ *        simultaneously
  *
- * Release both halves of a link simultaneously and deallocate
- * the structure.
+ * Release both halves of an external link simultaneously and 
+ * deallocate the structure.
  *
  * When the function is called it is assumed that:
  * - Neither of the parts of the link are part of any list or tree
@@ -2209,39 +2332,43 @@ erts_link_set_dead_dist(ErtsLink *lnk, Eterm nodename);
 Uint erts_link_size(ErtsLink *lnk);
 
 /* internal function... */
-void erts_link_destroy__(ErtsLinkData *ldp);
+void erts_link_destroy_elink__(ErtsELink *elnk);
 
 /* implementations for globally inlined link functions... */
 #if ERTS_GLB_INLINE_INCL_FUNC_DEF
 
 #ifdef ERTS_ML_DEBUG
-extern size_t erts_link_a_offset;
-extern size_t erts_link_b_offset;
+extern size_t erts_link_proc_offset;
+extern size_t erts_link_dist_offset;
 extern size_t erts_link_key_offset;
 #endif
 
-ERTS_GLB_INLINE ErtsLinkData *
-erts_link_to_data(ErtsLink *lnk)
+ERTS_GLB_INLINE ErtsELink *
+erts_link_to_elink(ErtsLink *lnk)
 {
-    ErtsLinkData *ldp = erts_ml_node_to_main_struct__((ErtsMonLnkNode *) lnk);
+    ErtsELink *elnk;
+
+    ERTS_ML_ASSERT(lnk->flags & ERTS_ML_FLG_EXTENDED);
+
+    elnk = (ErtsELink *) erts_ml_node_to_main_struct__((ErtsMonLnkNode *) lnk);
 
 #ifdef ERTS_ML_DEBUG
-    ERTS_ML_ASSERT(erts_link_a_offset == (size_t) ldp->a.offset);
-    ERTS_ML_ASSERT(erts_link_key_offset == (size_t) ldp->a.key_offset);
-    ERTS_ML_ASSERT(erts_link_b_offset == (size_t) ldp->b.offset);
-    ERTS_ML_ASSERT(erts_link_key_offset == (size_t) ldp->b.key_offset);
+    ERTS_ML_ASSERT(erts_link_proc_offset == (size_t) elnk->ld.proc.offset);
+    ERTS_ML_ASSERT(erts_link_key_offset == (size_t) elnk->ld.proc.key_offset);
+    ERTS_ML_ASSERT(erts_link_dist_offset == (size_t) elnk->ld.dist.offset);
+    ERTS_ML_ASSERT(erts_link_key_offset == (size_t) elnk->ld.dist.key_offset);
 #endif
 
-    return ldp;
+    return elnk;
 }
 
 ERTS_GLB_INLINE ErtsLink *
-erts_link_to_other(ErtsLink *lnk, ErtsLinkData **ldpp)
+erts_link_to_other(ErtsLink *lnk, ErtsELink **elnkpp)
 {
-    ErtsLinkData *ldp = erts_link_to_data(lnk);
-    if (ldpp)
-        *ldpp = ldp;
-    return lnk == &ldp->a ? &ldp->b : &ldp->a;
+    ErtsELink *elnk = erts_link_to_elink(lnk);
+    if (elnkpp)
+        *elnkpp = elnk;
+    return lnk == &elnk->ld.proc ? &elnk->ld.dist : &elnk->ld.proc;
 }
 
 ERTS_GLB_INLINE int
@@ -2274,24 +2401,39 @@ erts_link_list_last(ErtsLink *list)
     return (ErtsLink *) erts_ml_dl_list_last__((ErtsMonLnkNode *) list);
 }
 
+ERTS_GLB_INLINE void
+erts_link_internal_release(ErtsLink *lnk)
+{
+    ERTS_ML_ASSERT(lnk->type == ERTS_LNK_TYPE_PROC
+                   || lnk->type == ERTS_LNK_TYPE_PORT);
+    ERTS_ML_ASSERT(!(lnk->flags & ERTS_ML_FLG_EXTENDED));
+    erts_free(ERTS_ALC_T_LINK, lnk);
+}
+
 ERTS_GLB_INLINE void
 erts_link_release(ErtsLink *lnk)
 {
-    ErtsLinkData *ldp = erts_link_to_data(lnk);
-    ERTS_ML_ASSERT(!(lnk->flags & ERTS_ML_FLG_IN_TABLE));
-    ERTS_ML_ASSERT(erts_atomic32_read_nob(&ldp->refc) > 0);
-    if (erts_atomic32_dec_read_nob(&ldp->refc) == 0)
-        erts_link_destroy__(ldp);
+    if (!(lnk->flags & ERTS_ML_FLG_EXTENDED))
+        erts_link_internal_release(lnk);
+    else {
+        ErtsELink *elnk = erts_link_to_elink(lnk);
+        ERTS_ML_ASSERT(!(lnk->flags & ERTS_ML_FLG_IN_TABLE));
+        ERTS_ML_ASSERT(erts_atomic32_read_nob(&elnk->ld.refc) > 0);
+        if (erts_atomic32_dec_read_nob(&elnk->ld.refc) == 0)
+            erts_link_destroy_elink__(elnk);
+    }
 }
 
 ERTS_GLB_INLINE void
 erts_link_release_both(ErtsLinkData *ldp)
 {
-    ERTS_ML_ASSERT(!(ldp->a.flags & ERTS_ML_FLG_IN_TABLE));
-    ERTS_ML_ASSERT(!(ldp->b.flags & ERTS_ML_FLG_IN_TABLE));
+    ERTS_ML_ASSERT(!(ldp->proc.flags & ERTS_ML_FLG_IN_TABLE));
+    ERTS_ML_ASSERT(!(ldp->dist.flags & ERTS_ML_FLG_IN_TABLE));
     ERTS_ML_ASSERT(erts_atomic32_read_nob(&ldp->refc) >= 2);
+    ERTS_ML_ASSERT(ldp->proc.flags & ERTS_ML_FLG_EXTENDED);
+    ERTS_ML_ASSERT(ldp->dist.flags & ERTS_ML_FLG_EXTENDED);
     if (erts_atomic32_add_read_nob(&ldp->refc, (erts_aint32_t) -2) == 0)
-        erts_link_destroy__(ldp);
+        erts_link_destroy_elink__((ErtsELink *) ldp);
 }
 
 ERTS_GLB_INLINE ErtsLink *
@@ -2322,25 +2464,25 @@ erts_link_tree_key_delete(ErtsLink **root, ErtsLink *lnk)
 ERTS_GLB_INLINE int
 erts_link_dist_insert(ErtsLink *lnk, ErtsMonLnkDist *dist)
 {
-    ErtsLinkDataExtended *ldep;
+    ErtsELink *elnk;
     int insert;
 
     ERTS_ML_ASSERT(lnk->flags & ERTS_ML_FLG_EXTENDED);
     ERTS_ML_ASSERT(lnk->type == ERTS_LNK_TYPE_DIST_PROC);
 
-    ldep = (ErtsLinkDataExtended *) erts_link_to_data(lnk);
+    elnk = erts_link_to_elink(lnk);
 
-    ERTS_ML_ASSERT(!ldep->dist);
+    ERTS_ML_ASSERT(!elnk->dist);
     ERTS_ML_ASSERT(dist);
-    ldep->dist = dist;
-
-    erts_mon_link_dist_inc_refc(dist);
 
     erts_mtx_lock(&dist->mtx);
 
     insert = dist->alive;
-    if (insert)
+    if (insert) {
+        elnk->dist = dist;
+        erts_mon_link_dist_inc_refc(dist);
         erts_link_list_insert(&dist->links, lnk);
+    }
 
     erts_mtx_unlock(&dist->mtx);
 
@@ -2350,27 +2492,27 @@ erts_link_dist_insert(ErtsLink *lnk, ErtsMonLnkDist *dist)
 ERTS_GLB_INLINE int
 erts_link_dist_delete(ErtsLink *lnk)
 {
-    ErtsLinkDataExtended *ldep;
+    ErtsELink *elnk;
     ErtsMonLnkDist *dist;
-    int delete;
+    int delete_;
 
     ERTS_ML_ASSERT(lnk->flags & ERTS_ML_FLG_EXTENDED);
     ERTS_ML_ASSERT(lnk->type == ERTS_LNK_TYPE_DIST_PROC);
 
-    ldep = (ErtsLinkDataExtended *) erts_link_to_data(lnk);
-    dist = ldep->dist;
+    elnk = erts_link_to_elink(lnk);
+    dist = elnk->dist;
     if (!dist)
         return -1;
 
     erts_mtx_lock(&dist->mtx);
 
-    delete = !!dist->alive & !!(lnk->flags & ERTS_ML_FLG_IN_TABLE);
-    if (delete)
+    delete_ = !!dist->alive & !!(lnk->flags & ERTS_ML_FLG_IN_TABLE);
+    if (delete_)
         erts_link_list_delete(&dist->links, lnk);
 
     erts_mtx_unlock(&dist->mtx);
 
-    return delete;
+    return delete_;
 }
 
 
