@@ -1,7 +1,7 @@
 /*
  * %CopyrightBegin%
  *
- * Copyright Ericsson AB 2005-2020. All Rights Reserved.
+ * Copyright Ericsson AB 2005-2021. All Rights Reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -76,6 +76,22 @@ do {									\
     (CNT) += res__;							\
 } while (0)
 
+#define PRINT_UWORD64(CNT, FN, ARG, C, P, W, I)				\
+do {									\
+    int res__ = erts_printf_uword64((FN), (ARG), (C), (P), (W), (I));	\
+    if (res__ < 0)							\
+	return res__;							\
+    (CNT) += res__;							\
+} while (0)
+
+#define PRINT_SWORD64(CNT, FN, ARG, C, P, W, I)				\
+do {									\
+    int res__ = erts_printf_sword64((FN), (ARG), (C), (P), (W), (I));	\
+    if (res__ < 0)							\
+	return res__;							\
+    (CNT) += res__;							\
+} while (0)
+
 #define PRINT_DOUBLE(CNT, FN, ARG, C, P, W, I)				\
 do {									\
     int res__ = erts_printf_double((FN), (ARG), (C), (P), (W), (I));	\
@@ -84,6 +100,17 @@ do {									\
     (CNT) += res__;							\
 } while (0)
 
+#define PRINT_ATOM(CNT, FN, ARG, ATOM, DCOUNT)              \
+do {                                                        \
+    int res__ = print_atom_name(FN, ARG, ATOM, DCOUNT);     \
+    if (res__ < 0)                                          \
+        return res__;                                       \
+    (CNT) += res__;                                         \
+    if (*(DCOUNT) <= 0)                                     \
+        goto L_done;                                        \
+} while(0)
+
+
 /* CTYPE macros */
 
 #define LATIN1
@@ -112,8 +139,6 @@ do {									\
 #define IS_CNTRL(c)  ((c) < ' ' || (c) >= 127)
 #endif
 
-#define IS_PRINT(c)  (!IS_CNTRL(c))
-
 /* return 0 if list is not a non-empty flat list of printable characters */
 
 static int
@@ -158,7 +183,36 @@ static int is_printable_ascii(byte* bytep, Uint bytesize, Uint bitoffs)
     return 1;
 }
 
-/* print a atom doing what quoting is necessary */
+/*
+ * Helper function for print_atom_name(). Not generally useful.
+ */
+static ERTS_INLINE int latin1_char(int c1, int c2)
+{
+    if ((c1 & 0x80) == 0) {
+        /* Plain old 7-bit ASCII. */
+        return c1;
+    } else if ((c1 & 0xE0) == 0xC0) {
+        /* Unicode code points from 0x80 through 0x7FF. */
+        ASSERT((c2 & 0xC0) == 0x80);
+        return (c1 & 0x1F) << 6 | (c2 & 0x3F);
+    } else if ((c1 & 0xC0) == 0x80) {
+        /* A continutation byte in a utf8 sequence. Pretend that it is
+         * a character that is allowed in an atom. */
+        return 'a';
+    } else {
+        /* The start of a utf8 sequence comprising three or four
+         * bytes. Always needs quoting. */
+        return 0;
+    }
+}
+
+/*
+ * Print a atom, quoting it if necessary.
+ *
+ * Atoms are encoded in utf8. Since we have full control over creation
+ * of atoms, the utf8 encoding is always correct and there is no need
+ * to check for errors.
+ */
 static int print_atom_name(fmtfn_t fn, void* arg, Eterm atom, long *dcount)
 {
     int n, i;
@@ -168,6 +222,7 @@ static int print_atom_name(fmtfn_t fn, void* arg, Eterm atom, long *dcount)
     byte *s;
     byte *cpos;
     int c;
+    int lc;
 
     res = 0;
     i = atom_val(atom);
@@ -185,27 +240,47 @@ static int print_atom_name(fmtfn_t fn, void* arg, Eterm atom, long *dcount)
     *dcount -= atom_tab(i)->len;
 
     if (n == 0) {
+        /* The empty atom: '' */
 	PRINT_STRING(res, fn, arg, "''");
 	return res;
     }
 
+    /*
+     * Find out whether the atom will need quoting. Quoting is not necessary
+     * if the following applies:
+     *
+     *   - The first character is a lowercase letter in the Latin-1 code
+     *     block (0-255).
+     *
+     *   - All other characters are either alphanumeric characters in
+     *     the Latin-1 code block or the character '_'.
+     */
 
     need_quote = 0;
     cpos = s;
     pos = n - 1;
-
     c = *cpos++;
-    if (!IS_LOWER(c))
+    lc = latin1_char(c, *cpos);
+    if (!IS_LOWER(lc))
 	need_quote++;
     else {
 	while (pos--) {
 	    c = *cpos++;
-	    if (!IS_ALNUM(c) && (c != '_')) {
+            lc = latin1_char(c, *cpos);
+	    if (!IS_ALNUM(lc) && lc != '_') {
 		need_quote++;
 		break;
 	    }
 	}
     }
+
+    /*
+     * Now output the atom, including single quotes if needed.
+     *
+     * Control characters (including the range 128-159) must
+     * be specially printed. Therefore, we must do a partial
+     * decoding of the utf8 encoding.
+     */
     cpos = s;
     pos = n;
     if (need_quote)
@@ -222,12 +297,40 @@ static int print_atom_name(fmtfn_t fn, void* arg, Eterm atom, long *dcount)
 	case '\b': PRINT_STRING(res, fn, arg, "\\b"); break;
 	case '\v': PRINT_STRING(res, fn, arg, "\\v"); break;
 	default:
-	    if (IS_CNTRL(c)) {
+            if (c < ' ') {
+                /* ASCII control character (0-31). */
 		PRINT_CHAR(res, fn, arg, '\\');
 		PRINT_UWORD(res, fn, arg, 'o', 1, 3, (ErlPfUWord) c);
-	    }
-	    else
+            } else if (c >= 0x80) {
+                /* A multi-byte utf8-encoded code point. Determine the
+                 * length of the sequence. */
+                int n;
+                if ((c & 0xE0) == 0xC0) {
+                    n = 2;
+                } else if ((c & 0xF0) == 0xE0) {
+                    n = 3;
+                } else {
+                    ASSERT((c & 0xF8) == 0xF0);
+                    n = 4;
+                }
+                ASSERT(pos - n + 1 >= 0);
+
+                if (c == 0xC2 && *cpos < 0xA0) {
+                    /* Extended ASCII control character (128-159). */
+                    ASSERT(pos > 0);
+                    ASSERT(0x80 <= *cpos);
+                    PRINT_CHAR(res, fn, arg, '\\');
+                    PRINT_UWORD(res, fn, arg, 'o', 1, 3, (ErlPfUWord) *cpos);
+                    pos--, cpos++;
+                } else {
+                    PRINT_BUF(res, fn, arg, cpos-1, n);
+                    cpos += n - 1;
+                    pos -= n - 1;
+                }
+            } else {
+                /* Printable ASCII character. */
 		PRINT_CHAR(res, fn, arg, (char) c);
+            }
 	    break;
 	}
     }
@@ -236,7 +339,6 @@ static int print_atom_name(fmtfn_t fn, void* arg, Eterm atom, long *dcount)
     return res;
 }
 
-
 #define PRT_BAR                ((Eterm) 0)
 #define PRT_COMMA              ((Eterm) 1)
 #define PRT_CLOSE_LIST         ((Eterm) 2)
@@ -335,10 +437,29 @@ print_term(fmtfn_t fn, void* arg, Eterm obj, long *dcount) {
 	if ((*dcount)-- <= 0)
 	    goto L_done;
 
-	if (is_CP(obj)) {
-	    PRINT_STRING(res, fn, arg, "<cp/header:");
-	    PRINT_POINTER(res, fn, arg, cp_val(obj));
-	    PRINT_CHAR(res, fn, arg, '>');
+	if (is_non_value(obj)) {
+            PRINT_STRING(res, fn, arg, "<TNV>");
+	    goto L_done;
+        } else if (is_CP(obj)) {
+            const ErtsCodeMFA* mfa = erts_find_function_from_pc(cp_val(obj));
+            if (mfa) {
+                const UWord *func_start = erts_codemfa_to_code(mfa);
+                const UWord *cp_addr = (UWord*)cp_val(obj);
+
+                PRINT_STRING(res, fn, arg, "<");
+                PRINT_ATOM(res, fn, arg, mfa->module, dcount);
+                PRINT_STRING(res, fn, arg, ":");
+                PRINT_ATOM(res, fn, arg, mfa->function, dcount);
+                PRINT_STRING(res, fn, arg, "/");
+                PRINT_UWORD(res, fn, arg, 'u', 0, 1, (ErlPfUWord) mfa->arity);
+                PRINT_STRING(res, fn, arg, "+");
+                PRINT_UWORD(res, fn, arg, 'u', 0, 1, (ErlPfUWord) (cp_addr - func_start));
+                PRINT_STRING(res, fn, arg, ">");
+            } else {
+                PRINT_STRING(res, fn, arg, "<cp/header:");
+                PRINT_POINTER(res, fn, arg, cp_val(obj));
+                PRINT_CHAR(res, fn, arg, '>');
+            }
 	    goto L_done;
 	}
 	wobj = (Wterm)obj;
@@ -347,14 +468,7 @@ print_term(fmtfn_t fn, void* arg, Eterm obj, long *dcount) {
 	    PRINT_STRING(res, fn, arg, "[]");
 	    break;
 	case ATOM_DEF: {
-	    int tres = print_atom_name(fn, arg, obj, dcount);
-	    if (tres < 0) {
-		res = tres;
-		goto L_done;
-	    }
-	    res += tres;
-	    if (*dcount <= 0)
-		goto L_done;
+            PRINT_ATOM(res, fn, arg, obj, dcount);
 	    break;
 	}
 	case SMALL_DEF:
@@ -415,8 +529,8 @@ print_term(fmtfn_t fn, void* arg, Eterm obj, long *dcount) {
 	    PRINT_UWORD(res, fn, arg, 'u', 0, 1,
 			(ErlPfUWord) port_channel_no(wobj));
 	    PRINT_CHAR(res, fn, arg, '.');
-	    PRINT_UWORD(res, fn, arg, 'u', 0, 1,
-			(ErlPfUWord) port_number(wobj));
+	    PRINT_UWORD64(res, fn, arg, 'u', 0, 1,
+			  (ErlPfUWord64) port_number(wobj));
 	    PRINT_CHAR(res, fn, arg, '>');
 	    break;
 	case LIST_DEF:
@@ -582,14 +696,15 @@ print_term(fmtfn_t fn, void* arg, Eterm obj, long *dcount) {
 		PRINT_CHAR(res, fn, arg, '>');
 	    }
 	    break;
-	case MAP_DEF:
-            if (is_flatmap(wobj)) {
+	case MAP_DEF: {
+            Eterm *head = boxed_val(wobj);
+
+            if (is_flatmap_header(*head)) {
                 Uint n;
                 Eterm *ks, *vs;
-                flatmap_t *mp = (flatmap_t *)flatmap_val(wobj);
-                n  = flatmap_get_size(mp);
-                ks = flatmap_get_keys(mp);
-                vs = flatmap_get_values(mp);
+                n  = flatmap_get_size(head);
+                ks = flatmap_get_keys(head);
+                vs = flatmap_get_values(head);
 
                 PRINT_CHAR(res, fn, arg, '#');
                 PRINT_CHAR(res, fn, arg, '{');
@@ -604,8 +719,6 @@ print_term(fmtfn_t fn, void* arg, Eterm obj, long *dcount) {
                 }
             } else {
                 Uint n, mapval;
-                Eterm *head;
-                head = hashmap_val(wobj);
                 mapval = MAP_HEADER_VAL(*head);
                 switch (MAP_HEADER_TYPE(*head)) {
                 case MAP_HEADER_TAG_HAMT_HEAD_ARRAY:
@@ -650,6 +763,7 @@ print_term(fmtfn_t fn, void* arg, Eterm obj, long *dcount) {
                 }
             }
             break;
+        }
 	case MATCHSTATE_DEF:
 	    PRINT_STRING(res, fn, arg, "#MatchState");
 	    break;
