@@ -1,7 +1,7 @@
 /*
  * %CopyrightBegin%
  *
- * Copyright Ericsson AB 1996-2020. All Rights Reserved.
+ * Copyright Ericsson AB 1996-2021. All Rights Reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -59,7 +59,6 @@ struct enif_environment_t /* ErlNifEnv */
     Eterm* hp;
     Eterm* hp_end;
     ErlHeapFragment* heap_frag;
-    int fpe_was_unmasked;
     struct enif_tmp_obj_t* tmp_obj_list;
     int exception_thrown; /* boolean */
     Process *tracee;
@@ -109,7 +108,7 @@ typedef struct ErtsResource_
 
 extern Eterm erts_bld_resource_ref(Eterm** hp, ErlOffHeap*, ErtsResource*);
 
-extern BeamInstr* erts_call_nif_early(Process* c_p, ErtsCodeInfo* ci);
+extern ErtsCodePtr erts_call_nif_early(Process* c_p, const ErtsCodeInfo* ci);
 extern void erts_pre_nif(struct enif_environment_t*, Process*,
 			 struct erl_module_nif*, Process* tracee);
 extern void erts_post_nif(struct enif_environment_t* env);
@@ -124,7 +123,7 @@ extern Eterm erts_nif_taints(Process* p);
 extern void erts_print_nif_taints(fmtfn_t to, void* to_arg);
 
 /* Loads the specified NIF. The caller must have code write permission. */
-Eterm erts_load_nif(Process *c_p, BeamInstr *I, Eterm filename, Eterm args);
+Eterm erts_load_nif(Process *c_p, ErtsCodePtr I, Eterm filename, Eterm args);
 
 void erts_unload_nif(struct erl_module_nif* nif);
 extern void erl_nif_init(void);
@@ -137,7 +136,7 @@ extern Eterm erts_nif_call_function(Process *p, Process *tracee,
                                     int argc, Eterm *argv);
 
 int erts_call_dirty_nif(ErtsSchedulerData *esdp, Process *c_p,
-			BeamInstr *I, Eterm *reg);
+                        ErtsCodePtr I, Eterm *reg);
 ErtsMessage* erts_create_message_from_nif_env(ErlNifEnv* msg_env);
 
 
@@ -340,6 +339,15 @@ void erl_grow_estack(ErtsEStack*, Uint need);
 #define ESTK_CONCAT(a,b) a##b
 #define ESTK_DEF_STACK(s) ESTK_CONCAT(s,_default_estack)
 
+#define ESTACK_DEFAULT_VALUE(estack_default_stack_array, alloc_type)    \
+    (ErtsEStack) {                                                      \
+        estack_default_stack_array,  /* start */                        \
+        estack_default_stack_array,  /* sp */                           \
+        estack_default_stack_array + DEF_ESTACK_SIZE, /* end */         \
+        estack_default_stack_array,  /* default */                      \
+        alloc_type /* alloc_type */                                     \
+    }
+
 #define DECLARE_ESTACK(s)				\
     Eterm ESTK_DEF_STACK(s)[DEF_ESTACK_SIZE];		\
     ErtsEStack s = {					\
@@ -366,6 +374,30 @@ do {							\
     }							\
 } while(0)
 
+#define DESTROY_ESTACK_EXPLICIT_DEFAULT_ARRAY(s, the_estack_default_array) \
+    do {							\
+        if ((s).start != the_estack_default_array) {            \
+            erts_free((s).alloc_type, (s).start); 		\
+        }							\
+    } while(0)
+
+/* Allocate an array on the heap and move the stack there if the
+   default array (that is allocated on the heap is used) */
+#define ENSURE_ESTACK_HEAP_STACK_ARRAY(s, the_estack_default_array)\
+do {\
+    if ((s).start == the_estack_default_array) {\
+	UWord _wsz = ESTACK_COUNT(s);\
+        Eterm *_prev_stack_array = s.start;\
+	(s).start = erts_alloc((s).alloc_type,                          \
+			       DEF_ESTACK_SIZE * sizeof(Eterm));\
+	sys_memcpy((s).start, _prev_stack_array, _wsz*sizeof(Eterm));\
+	(s).sp = (s).start + _wsz;\
+	(s).end = (s).start + DEF_ESTACK_SIZE;\
+	(s).alloc_type = (s).alloc_type;\
+    }\
+    (s).edefault = NULL;\
+ } while (0)
+
 
 /*
  * Do not free the stack after this, it may have pointers into what
@@ -411,28 +443,29 @@ do {						\
 
 #define ESTACK_IS_STATIC(s) ((s).start == ESTK_DEF_STACK(s))
 
+#define ESTACK_RESERVE(s, push_cnt)             \
+do {					        \
+    if ((s).end - (s).sp < (Sint)(push_cnt)) {	\
+	erl_grow_estack(&(s), (push_cnt));	\
+    }					        \
+} while(0)
+
 #define ESTACK_PUSH(s, x)			\
 do {						\
-    if ((s).sp == (s).end) {			\
-	erl_grow_estack(&(s), 1); 		\
-    }						\
+    ESTACK_RESERVE(s, 1);                       \
     *(s).sp++ = (x);				\
 } while(0)
 
 #define ESTACK_PUSH2(s, x, y)			\
 do {						\
-    if ((s).sp > (s).end - 2) {			\
-	erl_grow_estack(&(s), 2);		\
-    }						\
+    ESTACK_RESERVE(s, 2);                       \
     *(s).sp++ = (x);				\
     *(s).sp++ = (y);				\
 } while(0)
 
 #define ESTACK_PUSH3(s, x, y, z)		\
 do {						\
-    if ((s).sp > (s).end - 3) {			\
-	erl_grow_estack(&s, 3); 		\
-    }						\
+    ESTACK_RESERVE(s, 3);                       \
     *(s).sp++ = (x);				\
     *(s).sp++ = (y);				\
     *(s).sp++ = (z);				\
@@ -440,22 +473,13 @@ do {						\
 
 #define ESTACK_PUSH4(s, E1, E2, E3, E4)		\
 do {						\
-    if ((s).sp > (s).end - 4) {			\
-	erl_grow_estack(&s, 4);                 \
-    }						\
+    ESTACK_RESERVE(s, 4);                       \
     *(s).sp++ = (E1);				\
     *(s).sp++ = (E2);				\
     *(s).sp++ = (E3);				\
     *(s).sp++ = (E4);				\
 } while(0)
 
-#define ESTACK_RESERVE(s, push_cnt)             \
-do {					        \
-    if ((s).sp > (s).end - (push_cnt)) {	\
-	erl_grow_estack(&(s), (push_cnt));	\
-    }					        \
-} while(0)
-
 /* Must be preceded by ESTACK_RESERVE */
 #define ESTACK_FAST_PUSH(s, x)				\
 do {							\
@@ -465,7 +489,7 @@ do {							\
 
 #define ESTACK_COUNT(s) ((s).sp - (s).start)
 #define ESTACK_ISEMPTY(s) ((s).sp == (s).start)
-#define ESTACK_POP(s) (*(--(s).sp))
+#define ESTACK_POP(s) (ASSERT(!ESTACK_ISEMPTY(s)),(*(--(s).sp)))
 
 
 /*
@@ -486,6 +510,15 @@ void erl_grow_wstack(ErtsWStack*, Uint need);
 #define WSTK_CONCAT(a,b) a##b
 #define WSTK_DEF_STACK(s) WSTK_CONCAT(s,_default_wstack)
 
+#define WSTACK_DEFAULT_VALUE(wstack_default_stack_array, alloc_type)    \
+    (ErtsWStack) {                                                      \
+        wstack_default_stack_array,  /* start */                        \
+        wstack_default_stack_array,  /* sp */                           \
+        wstack_default_stack_array + DEF_ESTACK_SIZE, /* end */         \
+        wstack_default_stack_array,  /* default */                      \
+        alloc_type /* alloc_type */                                     \
+    }
+
 #define WSTACK_DECLARE(s)				\
     UWord WSTK_DEF_STACK(s)[DEF_WSTACK_SIZE];		\
     ErtsWStack s = {					\
@@ -528,6 +561,28 @@ do {							\
 } while(0)
 #define DESTROY_WSTACK WSTACK_DESTROY
 
+#define DESTROY_WSTACK_EXPLICIT_DEFAULT_ARRAY(s, the_wstack_default_array) \
+    do {                                                                \
+        if ((s).wstart != the_wstack_default_array) {                   \
+            erts_free((s).alloc_type, (s).wstart);                      \
+        }                                                               \
+    } while(0)
+
+#define ENSURE_WSTACK_HEAP_STACK_ARRAY(s, the_wstack_default_array)\
+do {\
+    if ((s).wstart == the_wstack_default_array) {\
+	UWord _wsz = WSTACK_COUNT(s);\
+        UWord *_prev_stack_array = s.wstart;\
+	(s).wstart = erts_alloc((s).alloc_type,                          \
+                                DEF_WSTACK_SIZE * sizeof(UWord));       \
+	sys_memcpy((s).wstart, _prev_stack_array, _wsz*sizeof(UWord));\
+	(s).wsp = (s).wstart + _wsz;\
+	(s).wend = (s).wstart + DEF_WSTACK_SIZE;\
+	(s).alloc_type = (s).alloc_type;\
+    }\
+    (s).wdefault = NULL;\
+ } while (0)
+
 #define WSTACK_DEBUG(s) \
     do { \
 	fprintf(stderr, "wstack size   = %ld\r\n", s.wsp - s.wstart); \
@@ -579,28 +634,29 @@ do {						\
 
 #define WSTACK_IS_STATIC(s) (s.wstart == WSTK_DEF_STACK(s))
 
-#define WSTACK_PUSH(s, x)				\
-do {							\
-    if (s.wsp == s.wend) {				\
-	erl_grow_wstack(&s, 1); 	                \
-    }							\
-    *s.wsp++ = (x);					\
+#define WSTACK_RESERVE(s, push_cnt)             \
+do {						\
+    if (s.wend - s.wsp < (Sint)(push_cnt)) {    \
+	erl_grow_wstack(&s, (push_cnt));        \
+    }                                           \
+} while(0)
+
+#define WSTACK_PUSH(s, x)                       \
+do {                                            \
+    WSTACK_RESERVE(s, 1);                       \
+    *s.wsp++ = (x);				\
 } while(0)
 
 #define WSTACK_PUSH2(s, x, y)			\
 do {						\
-    if (s.wsp > s.wend - 2) {			\
-	erl_grow_wstack(&s, 2);                 \
-    }						\
+    WSTACK_RESERVE(s, 2);                       \
     *s.wsp++ = (x);				\
     *s.wsp++ = (y);				\
 } while(0)
 
 #define WSTACK_PUSH3(s, x, y, z)		\
 do {						\
-    if (s.wsp > s.wend - 3) {	                \
-	erl_grow_wstack(&s, 3);                 \
-    }						\
+    WSTACK_RESERVE(s, 3);                       \
     *s.wsp++ = (x);				\
     *s.wsp++ = (y);				\
     *s.wsp++ = (z);				\
@@ -608,9 +664,7 @@ do {						\
 
 #define WSTACK_PUSH4(s, A1, A2, A3, A4)		\
 do {						\
-    if (s.wsp > s.wend - 4) {	                \
-	erl_grow_wstack(&s, 4);                 \
-    }						\
+    WSTACK_RESERVE(s, 4);                       \
     *s.wsp++ = (A1);				\
     *s.wsp++ = (A2);				\
     *s.wsp++ = (A3);				\
@@ -619,9 +673,7 @@ do {						\
 
 #define WSTACK_PUSH5(s, A1, A2, A3, A4, A5)     \
 do {						\
-    if (s.wsp > s.wend - 5) {	                \
-	erl_grow_wstack(&s, 5);                 \
-    }						\
+    WSTACK_RESERVE(s, 5);                       \
     *s.wsp++ = (A1);				\
     *s.wsp++ = (A2);				\
     *s.wsp++ = (A3);				\
@@ -631,9 +683,7 @@ do {						\
 
 #define WSTACK_PUSH6(s, A1, A2, A3, A4, A5, A6) \
 do {						\
-    if (s.wsp > s.wend - 6) {	                \
-	erl_grow_wstack(&s, 6);                 \
-    }						\
+    WSTACK_RESERVE(s, 6);                       \
     *s.wsp++ = (A1);				\
     *s.wsp++ = (A2);				\
     *s.wsp++ = (A3);				\
@@ -642,13 +692,6 @@ do {						\
     *s.wsp++ = (A6);				\
 } while(0)
 
-#define WSTACK_RESERVE(s, push_cnt)             \
-do {						\
-    if (s.wsp > s.wend - (push_cnt)) { 	        \
-	erl_grow_wstack(&s, (push_cnt));        \
-    }                                           \
-} while(0)
-
 /* Must be preceded by WSTACK_RESERVE */
 #define WSTACK_FAST_PUSH(s, x)                  \
 do {						\
@@ -868,7 +911,7 @@ ERTS_GLB_INLINE Eterm erts_equeue_get(ErtsEQueue *q) {
 void erts_emasculate_writable_binary(ProcBin* pb);
 Eterm erts_new_heap_binary(Process *p, byte *buf, int len, byte** datap);
 Eterm erts_new_mso_binary(Process*, byte*, Uint);
-Eterm new_binary(Process*, byte*, Uint);
+Eterm new_binary(Process*, const byte*, Uint);
 Eterm erts_heap_factory_new_binary(ErtsHeapFactory *hfact, byte *buf,
                                    Uint len, Uint reserve_size);
 Eterm erts_realloc_binary(Eterm bin, size_t size);
@@ -889,8 +932,6 @@ void erts_bif_info_init(void);
 
 /* bif.c */
 
-void erts_write_bif_wrapper(Export *export, BeamInstr *address);
-
 void erts_queue_monitor_message(Process *,
 				ErtsProcLocks*,
 				Eterm,
@@ -898,7 +939,7 @@ void erts_queue_monitor_message(Process *,
 				Eterm,
 				Eterm);
 void erts_init_trap_export(Export* ep, Eterm m, Eterm f, Uint a,
-			   Eterm (*bif)(Process*, Eterm*, BeamInstr*));
+			   Eterm (*bif)(Process*, Eterm*, ErtsCodePtr));
 void erts_init_bif(void);
 Eterm erl_send(Process *p, Eterm to, Eterm msg);
 int erts_set_group_leader(Process *proc, Eterm new_gl);
@@ -914,7 +955,11 @@ Eterm erl_is_function(Process* p, Eterm arg1, Eterm arg2);
 
 /* beam_bif_load.c */
 Eterm erts_check_process_code(Process *c_p, Eterm module, int *redsp, int fcalls);
-Eterm erts_proc_copy_literal_area(Process *c_p, int *redsp, int fcalls, int gc_allowed);
+#define ERTS_CLA_SCAN_WORDS_PER_RED 512
+
+int erts_check_copy_literals_gc_need(Process *c_p, int *redsp,
+                                     char *literals, Uint lit_bsize);
+Eterm erts_copy_literals_gc(Process *c_p, int *redsp, int fcalls);
 
 Uint32 erts_block_release_literal_area(void);
 void erts_unblock_release_literal_area(Uint32);
@@ -944,15 +989,22 @@ extern Process *erts_code_purger;
 
 /* beam_load.c */
 typedef struct {
-    ErtsCodeMFA* mfa;		/* Pointer to: Mod, Name, Arity */
-    Uint needed;		/* Heap space needed for entire tuple */
-    Uint32 loc;			/* Location in source code */
-    Eterm* fname_ptr;		/* Pointer to fname table */
+    const ErtsCodeMFA* mfa;     /* Pointer to: Mod, Name, Arity */
+    Uint needed;                /* Heap space needed for entire tuple */
+    Uint32 loc;                 /* Location in source code */
+    const Eterm* fname_ptr;     /* Pointer to fname table */
 } FunctionInfo;
 
 Binary* erts_alloc_loader_state(void);
+
+/* Return the module name (a tagged atom) for the prepared code in the magic
+ * binary, or NIL if the binary does not contain prepared code. */
 Eterm erts_module_for_prepared_code(Binary* magic);
+
+/* Return a non-zero value if the prepared module has an on_load function,
+ * or 0 if it does not. */
 Eterm erts_has_code_on_load(Binary* magic);
+
 Eterm erts_prepare_loading(Binary* loader_state,  Process *c_p,
 			   Eterm group_leader, Eterm* modp,
 			   byte* code, Uint size);
@@ -961,23 +1013,22 @@ Eterm erts_finish_loading(Binary* loader_state, Process* c_p,
 Eterm erts_preload_module(Process *c_p, ErtsProcLocks c_p_locks,
 			  Eterm group_leader, Eterm* mod, byte* code, Uint size);
 void init_load(void);
-ErtsCodeMFA* find_function_from_pc(BeamInstr* pc);
+const ErtsCodeMFA* erts_find_function_from_pc(ErtsCodePtr pc);
 Eterm* erts_build_mfa_item(FunctionInfo* fi, Eterm* hp,
-			   Eterm args, Eterm* mfa_p);
-void erts_set_current_function(FunctionInfo* fi, ErtsCodeMFA* mfa);
-Eterm erts_module_info_0(Process* p, Eterm module);
-Eterm erts_module_info_1(Process* p, Eterm module, Eterm what);
+			   Eterm args, Eterm* mfa_p, Eterm loc_tail);
+void erts_set_current_function(FunctionInfo* fi, const ErtsCodeMFA* mfa);
 Eterm erts_make_stub_module(Process* p, Eterm Mod, Eterm Beam, Eterm Info);
-int erts_commit_hipe_patch_load(Eterm hipe_magic_bin);
 
 /* beam_ranges.c */
 void erts_init_ranges(void);
 void erts_start_staging_ranges(int num_new);
 void erts_end_staging_ranges(int commit);
-void erts_update_ranges(BeamInstr* code, Uint size);
-void erts_remove_from_ranges(BeamInstr* code);
+void erts_update_ranges(const BeamCodeHeader* code, Uint size);
+void erts_remove_from_ranges(const BeamCodeHeader* code);
 UWord erts_ranges_sz(void);
-void erts_lookup_function_info(FunctionInfo* fi, BeamInstr* pc, int full_info);
+void erts_lookup_function_info(FunctionInfo* fi,
+                               ErtsCodePtr pc,
+                               int full_info);
 extern ErtsLiteralArea** erts_dump_lit_areas;
 extern Uint erts_dump_num_lit_areas;
 
@@ -989,7 +1040,7 @@ void process_info(fmtfn_t, void *);
 void print_process_info(fmtfn_t, void *, Process*, ErtsProcLocks);
 void info(fmtfn_t, void *);
 void loaded(fmtfn_t, void *);
-void erts_print_base64(fmtfn_t to, void *to_arg, byte* src, Uint size);
+void erts_print_base64(fmtfn_t to, void *to_arg, const byte* src, Uint size);
 
 /* sighandler sys.c */
 int erts_set_signal(Eterm signal, Eterm type);
@@ -1000,9 +1051,9 @@ double erts_get_positive_zero_float(void);
 /* config.c */
 
 __decl_noreturn void __noreturn erts_exit_epilogue(void);
-__decl_noreturn void __noreturn erts_exit(int n, char*, ...);
+__decl_noreturn void __noreturn erts_exit(int n, const char*, ...);
 __decl_noreturn void __noreturn erts_flush_async_exit(int n, char*, ...);
-void erl_error(char*, va_list);
+void erl_error(const char*, va_list);
 
 /* This controls whether sharing-preserving copy is used by Erlang */
 
@@ -1020,6 +1071,9 @@ typedef struct {
     ErtsAlcType_t queue_alloc_type;
     UWord  bitstore_default[DEF_WSTACK_SIZE];
     UWord* bitstore_start;
+#ifdef DEBUG
+    UWord* bitstore_stop;
+#endif
     ErtsAlcType_t bitstore_alloc_type;
     Eterm  shtable_default[DEF_ESTACK_SIZE];
     Eterm* shtable_start;
@@ -1161,12 +1215,12 @@ void print_pass_through(int, byte*, int);
 /* beam_emu.c */
 int catchlevel(Process*);
 void init_emulator(void);
-void process_main(Eterm* x_reg_array, FloatDef* f_reg_array);
+void process_main(ErtsSchedulerData *);
 void erts_dirty_process_main(ErtsSchedulerData *);
 Eterm build_stacktrace(Process* c_p, Eterm exc);
 Eterm expand_error_value(Process* c_p, Uint freason, Eterm Value);
 void erts_save_stacktrace(Process* p, struct StackTrace* s, int depth);
-BeamInstr *erts_printable_return_address(Process* p, Eterm *E) ERTS_NOINLINE;
+ErtsCodePtr erts_printable_return_address(Process* p, Eterm *E) ERTS_NOINLINE;
 
 /* erl_init.c */
 
@@ -1191,6 +1245,11 @@ extern int erts_no_line_info;
 extern Eterm erts_error_logger_warnings;
 extern int erts_initialized;
 extern int erts_compat_rel;
+
+#ifdef BEAMASM
+extern int erts_jit_asm_dump;
+#endif
+
 void erl_start(int, char**);
 void erts_usage(void);
 Eterm erts_preloaded(Process* p);
@@ -1269,15 +1328,94 @@ Uint64 erts_timestamp_millis(void);
 Export* erts_find_function(Eterm, Eterm, unsigned int, ErtsCodeIndex);
 
 /* ERTS_NOINLINE prevents link-time optimization across modules */
-void *erts_calc_stacklimit(char *prev_c, UWord stacksize) ERTS_NOINLINE;
+const void *erts_get_stacklimit(void);
 int erts_check_below_limit(char *ptr, char *limit) ERTS_NOINLINE;
 int erts_check_above_limit(char *ptr, char *limit) ERTS_NOINLINE;
+
 void *erts_ptr_id(void *ptr) ERTS_NOINLINE;
 int erts_check_if_stack_grows_downwards(char *ptr) ERTS_NOINLINE;
 
 Eterm store_external_or_ref_in_proc_(Process *, Eterm);
 Eterm store_external_or_ref_(Uint **, ErlOffHeap*, Eterm);
 
+typedef Eterm  (*erts_ycf_continue_fun_t)(long* ycf_number_of_reduction_param,
+                                          void** ycf_trap_state,
+                                          void* ycf_extra_context);
+typedef void (*erts_ycf_destroy_trap_state_fun_t)(void *trap_state);
+typedef Eterm (*erts_ycf_yielding_fun_t)(long* ycf_nr_of_reductions_param,
+                                         void** ycf_trap_state,
+                                         void* ycf_extra_context,
+                                         void* (*ycf_yield_alloc_fun) (size_t,void*),
+                                         void (*ycf_yield_free_fun) (void*,void*),
+                                         void* ycf_yield_alloc_free_context,
+                                         size_t ycf_stack_alloc_size_or_max_size,
+                                         void* ycf_stack_alloc_data,
+                                         Process* p,
+                                         Eterm* bif_args);
+Eterm erts_ycf_trap_driver(Process* p,
+                           Eterm* bif_args,
+                           int nr_of_arguments,
+                           int iterations_per_red,
+                           ErtsAlcType_t memory_allocation_type,
+                           size_t ycf_stack_alloc_size,
+                           int export_entry_index,
+                           erts_ycf_continue_fun_t ycf_continue_fun,
+                           erts_ycf_destroy_trap_state_fun_t ycf_destroy_fun,
+                           erts_ycf_yielding_fun_t ycf_yielding_fun);
+
+/* A quick sort function that is compatible with the qsort function
+   declared in stdlib.h. We need our own so that we can yield inside
+   the function */
+typedef int (*erts_void_ptr_cmp_t)(const void *, const void *);
+void erts_qsort(void *base,
+                size_t nr_of_items,
+                size_t item_size,
+                erts_void_ptr_cmp_t compare);
+/* YCF generated functions for yielding of erts_qsort. This means that
+   the following three functions can be used when one needs a yieldable
+   sorting function. See
+   $ERL_TOP/erts/emulator/internal_doc/AutomaticYieldingOfCCode.md for
+   information about using YCF generated functions.
+
+   
+   !!!!
+   Note that the erts_qsort_swap that is used by erts_qsort does
+   not have yielding enabled. If the array items are large erts_qsort
+   should also trap in the erts_qsort_swap function, but this causes
+   terrible performance when the array items are small, so one should
+   investigate a fast-path approach
+*/
+void erts_qsort_ycf_gen_destroy(void* ycf_my_trap_state);
+void  erts_qsort_ycf_gen_continue(long* ycf_number_of_reduction_param,
+                                  void** ycf_trap_state,
+                                  void* ycf_extra_context /* Not used, can be NULL */);
+void erts_qsort_ycf_gen_yielding(long* ycf_nr_of_reductions_param,
+                                 void** ycf_trap_state,
+                                 void* ycf_extra_context, /* Not used, can be NULL */
+                                 void* (*ycf_yield_alloc_fun) (size_t,void*),
+                                 void (*ycf_yield_free_fun) (void*,void*),
+                                 void* ycf_yield_alloc_free_context,
+                                 size_t ycf_stack_alloc_size_or_max_size,  /* Not used, can be 0 */
+                                 void* ycf_stack_alloc_data, /* Not used, can be NULL */
+                                 void *base,
+                                 size_t nr_of_items,
+                                 size_t item_size,
+                                 erts_void_ptr_cmp_t compare);
+#if defined(DEBUG)
+/*
+ * ycf_debug_get_stack_start is used in YCF's debug mode (see
+ * documentation for the -debug flag of the YCF tool). The function
+ * ycf_debug_set_stack_start sets the value that the function
+ * ycf_debug_get_stack_start returns for the current thread. The
+ * function ycf_debug_reset_stack_start sets the value that the
+ * function ycf_debug_get_stack_start returns to NULL for the current
+ * thread.
+ */
+void ycf_debug_set_stack_start(void * start);
+void ycf_debug_reset_stack_start(void);
+void *ycf_debug_get_stack_start(void);
+#endif
+
 #define NC_HEAP_SIZE(NC) \
  (ASSERT(is_node_container((NC))), \
   IS_CONST((NC)) ? 0 : (thing_arityval(*boxed_val((NC))) + 1))
@@ -1313,6 +1451,7 @@ void erts_debug_foreach_persistent_term_off_heap(void (*func)(ErlOffHeap *, void
                                                  void *arg);
 int erts_debug_have_accessed_literal_area(ErtsLiteralArea *lap);
 void erts_debug_save_accessed_literal_area(ErtsLiteralArea *lap);
+Eterm erts_debug_persistent_term_xtra_info(Process* c_p);
 
 /* external.c */
 void erts_init_external(void);
@@ -1321,8 +1460,8 @@ void erts_init_external(void);
 void erts_init_map(void);
 
 /* beam_debug.c */
-UWord erts_check_stack_recursion_downwards(char *start_c);
-UWord erts_check_stack_recursion_upwards(char *start_c);
+UWord erts_check_stack_recursion_downwards(char *start_c, char *prev_c);
+UWord erts_check_stack_recursion_upwards(char *start_c, char *prev_c);
 int erts_is_above_stack_limit(char *ptr);
 
 /* erl_unicode.c */
@@ -1438,16 +1577,21 @@ Eterm collect_memory(Process *);
 void dump_memory_to_fd(int);
 int dump_memory_data(const char *);
 
+Eterm erts_unary_minus(Process* p, Eterm arg1);
 Eterm erts_mixed_plus(Process* p, Eterm arg1, Eterm arg2);
 Eterm erts_mixed_minus(Process* p, Eterm arg1, Eterm arg2);
 Eterm erts_mixed_times(Process* p, Eterm arg1, Eterm arg2);
 Eterm erts_mixed_div(Process* p, Eterm arg1, Eterm arg2);
+
+int erts_int_div_rem(Process* p, Eterm arg1, Eterm arg2, Eterm *q, Eterm *r);
 Eterm erts_int_div(Process* p, Eterm arg1, Eterm arg2);
 Eterm erts_int_rem(Process* p, Eterm arg1, Eterm arg2);
-Eterm erts_band(Process* p, Eterm arg1, Eterm arg2);
-Eterm erts_bor(Process* p, Eterm arg1, Eterm arg2);
 Eterm erts_bxor(Process* p, Eterm arg1, Eterm arg2);
+Eterm erts_bsr(Process* p, Eterm arg1, Eterm arg2);
+Eterm erts_bsl(Process* p, Eterm arg1, Eterm arg2);
 Eterm erts_bnot(Process* p, Eterm arg);
+Eterm erts_bor(Process* p, Eterm arg1, Eterm arg2);
+Eterm erts_band(Process* p, Eterm arg1, Eterm arg2);
 
 Eterm erts_gc_mixed_plus(Process* p, Eterm* reg, Uint live);
 Eterm erts_gc_mixed_minus(Process* p, Eterm* reg, Uint live);
@@ -1584,7 +1728,7 @@ dtrace_pid_str(Eterm pid, char *process_buf)
                       pid_number(pid),
                       pid_serial(pid));
     else if (is_port(pid))
-        erts_snprintf(process_buf, DTRACE_TERM_BUF_SIZE, "#Port<%lu.%lu>",
+        erts_snprintf(process_buf, DTRACE_TERM_BUF_SIZE, "#Port<%lu.%b64u>",
                       port_channel_no(pid),
                       port_number(pid));
 }
