@@ -1,7 +1,7 @@
 /*
  * %CopyrightBegin%
  * 
- * Copyright Ericsson AB 1997-2016. All Rights Reserved.
+ * Copyright Ericsson AB 1997-2021. All Rights Reserved.
  * 
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -157,7 +157,7 @@ static RECT winPos;
 static BOOL toolbarVisible;
 static BOOL destroyed = FALSE;
 
-static int lines_to_save = 1000; /* Maximum number of screen lines to save. */
+static int lines_to_save = 10000; /* Maximum number of screen lines to save. */
 
 #define TITLE_BUF_SZ 256
 
@@ -205,6 +205,18 @@ static void window_title(struct title_buf *);
 static void free_window_title(struct title_buf *);
 static void Client_OnMouseMove(HWND hwnd, int x, int y, UINT keyFlags);
 
+#ifdef HARDDEBUG
+/* For really hard GUI startup debugging, place DEBUGBOX() macros in code
+   and get modal message boxes with the line number. */
+static void debug_box(int line) {
+  TCHAR buff[1024];
+  swprintf(buff,1024,TEXT("DBG:%d"),line);
+  MessageBox(NULL,buff,TEXT("DBG"),MB_OK|MB_APPLMODAL);
+}
+
+#define DEBUGBOX() debug_box(__LINE__)
+#endif
+
 #define CON_VPRINTF_BUF_INC_SIZE 1024
 
 static erts_dsprintf_buf_t *
@@ -430,6 +442,13 @@ ConThreadInit(LPVOID param)
     struct title_buf title;
 
     /*DebugBreak();*/
+#ifdef HARDDEBUG
+    if(AttachConsole(ATTACH_PARENT_PROCESS) || AllocConsole()) {
+        freopen("CONOUT$", "w", stdout);
+        freopen("CONOUT$", "w", stderr);
+    }
+#endif
+
     hInstance = GetModuleHandle(NULL);
     StartupInfo.dwFlags = 0;
     GetStartupInfo(&StartupInfo);
@@ -549,7 +568,7 @@ FrameWndProc(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
     case WM_CREATE:
         /* client window creation */
 	window_title(&title);
-        hClientWnd = CreateWindowEx(WS_EX_CLIENTEDGE, szClientClass, title.name,
+        hClientWnd = CreateWindowEx(0, szClientClass, title.name,
 				    WS_CHILD|WS_VISIBLE|WS_VSCROLL|WS_HSCROLL,
 				    CW_USEDEFAULT, CW_USEDEFAULT,
 				    CW_USEDEFAULT, CW_USEDEFAULT,		
@@ -610,7 +629,7 @@ FrameWndProc(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
 		lpttt->hinst = hInstance;
 		/* check for combobox handle */
 		if (lpttt->uFlags&TTF_IDISHWND) {
-		    if ((lpttt->hdr.idFrom == (UINT) hComboWnd)) {
+		    if ((lpttt->hdr.idFrom == (UINT_PTR) hComboWnd)) {
 			lstrcpy(lpttt->lpszText,TEXT("Command History"));
 			break;
 		    }
@@ -1313,13 +1332,21 @@ LoadUserPreferences(void)
     DWORD size;
     DWORD res;
     DWORD type;
-
+    HFONT hfont;
     /* default prefs */
-    GetObject(GetStockObject(SYSTEM_FIXED_FONT),sizeof(LOGFONT),(PSTR)&logfont);
+    hfont = CreateFont(0,0, 0,0, 0, FALSE,FALSE,FALSE,
+                       ANSI_CHARSET, OUT_TT_ONLY_PRECIS, CLIP_DEFAULT_PRECIS,
+                       CLEARTYPE_QUALITY, FIXED_PITCH, TEXT("Consolas"));
+    if(hfont) {
+        GetObject(hfont, sizeof(LOGFONT), (PSTR)&logfont);
+        DeleteObject(hfont);
+    } else {
+        GetObject(GetStockObject(SYSTEM_FIXED_FONT),sizeof(LOGFONT),(PSTR)&logfont);
+    }
     fgColor = GetSysColor(COLOR_WINDOWTEXT);
     bkgColor = GetSysColor(COLOR_WINDOW);
     winPos.left = -1;
-    toolbarVisible = TRUE;
+    toolbarVisible = FALSE;
 
     if (RegCreateKeyEx(HKEY_CURRENT_USER, USER_KEY, 0, 0,
 		       REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL,
@@ -1774,7 +1801,7 @@ void ConChooseColor(HWND hwnd)
         SetBkColor(hdc,bkgColor);
         ReleaseDC(hwnd,hdc);
         hbrush = CreateSolidBrush(bkgColor);
-        DeleteObject((HBRUSH)SetClassLong(hClientWnd,GCL_HBRBACKGROUND,(LONG)hbrush));
+        DeleteObject((HBRUSH)SetClassLongPtr(hClientWnd,GCL_HBRBACKGROUND,(LONG_PTR)hbrush));
         InvalidateRect(hwnd,NULL,TRUE);		
     }    
 }
@@ -1972,9 +1999,11 @@ DIALOG_PROC_RET CALLBACK AboutDlgProc(HWND hDlg, UINT iMsg, WPARAM wParam, LPARA
         OffsetRect(&rc, -rc.left, -rc.top); 
         OffsetRect(&rc, -rcDlg.right, -rcDlg.bottom); 
         SetWindowPos(hDlg,HWND_TOP,rcOwner.left + (rc.right / 2), 
-		     rcOwner.top + (rc.bottom / 2),0,0,SWP_NOSIZE); 
-        SetDlgItemText(hDlg, ID_VERSIONSTRING,
-		       TEXT("Erlang emulator version ") TEXT(ERLANG_VERSION)); 
+		     rcOwner.top + (rc.bottom / 2),0,0,SWP_NOSIZE);
+        SetDlgItemText(hDlg, ID_OTP_VERSIONSTRING,
+		       TEXT("OTP version ") TEXT(ERLANG_OTP_VERSION));
+        SetDlgItemText(hDlg, ID_ERTS_VERSIONSTRING,
+		       TEXT("Erlang emulator version ") TEXT(ERLANG_VERSION));
         return (DIALOG_PROC_RET) TRUE;
     case WM_COMMAND:
         switch (LOWORD(wParam)) {
@@ -2012,7 +2041,7 @@ ConDrawText(HWND hwnd)
 	TCHAR *bu = (TCHAR *) ALLOC((num_chars+1) * sizeof(TCHAR));
 	memcpy(bu,buf,num_chars * sizeof(TCHAR));
 	bu[num_chars]='\0';
-	fprintf(stderr,TEXT("ConDrawText\"%s\"\n"),bu);
+	fprintf(stderr,"ConDrawText\"%S\"\n",bu);
 	FREE(bu);
 	fflush(stderr);
     }
@@ -2211,17 +2240,6 @@ static TBADDBITMAP tbbitmap =
     HINST_COMMCTRL, IDB_STD_SMALL_COLOR,
 };
 
-#ifdef HARDDEBUG
-/* For really hard GUI startup debugging, place DEBUGBOX() macros in code
-   and get modal message boxes with the line number. */
-static void debug_box(int line) {
-  TCHAR buff[1024];
-  swprintf(buff,1024,TEXT("DBG:%d"),line);
-  MessageBox(NULL,buff,TEXT("DBG"),MB_OK|MB_APPLMODAL);
-}
-
-#define DEBUGBOX() debug_box(__LINE__)  
-#endif
 
 static HWND
 InitToolBar(HWND hwndParent) 
@@ -2243,7 +2261,7 @@ InitToolBar(HWND hwndParent)
     SendMessage(hwndTB,TB_BUTTONSTRUCTSIZE,
 		(WPARAM) sizeof(TBBUTTON),0); 
     tbbitmap.hInst = NULL;
-    tbbitmap.nID   = (UINT) CreateMappedBitmap(beam_module, 1,0, &colorMap, 1);
+    tbbitmap.nID   = (UINT_PTR) CreateMappedBitmap(beam_module, 1,0, &colorMap, 1);
     SendMessage(hwndTB, TB_ADDBITMAP, (WPARAM) 4, 
 		(LPARAM) &tbbitmap); 
 
@@ -2269,7 +2287,7 @@ InitToolBar(HWND hwndParent)
     ti.cbSize = sizeof(TOOLINFO);
     ti.uFlags = TTF_IDISHWND|TTF_CENTERTIP|TTF_SUBCLASS;
     ti.hwnd = hwndTB;;
-    ti.uId = (UINT)hComboWnd;
+    ti.uId = (UINT_PTR)hComboWnd;
     ti.lpszText = LPSTR_TEXTCALLBACK;
     hwndTT = (HWND)SendMessage(hwndTB,TB_GETTOOLTIPS,0,0);
     SendMessage(hwndTT,TTM_ADDTOOL,0,(LPARAM)&ti);
