@@ -1,7 +1,7 @@
 /*
  * %CopyrightBegin%
  *
- * Copyright Ericsson AB 2018-2019. All Rights Reserved.
+ * Copyright Ericsson AB 2018-2021. All Rights Reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -65,6 +65,14 @@
 
 #ifdef HAVE_IFADDRS_H
 #include <ifaddrs.h>
+#elif defined(__PASE__)
+/* PASE has this, but under a different name because AIX doesn't have it. */
+#include <as400_protos.h>
+/*
+ * We don't redefine the function names because they're used in other
+ * contexts, but the struct is safe to rename.
+ */
+#define ifaddrs ifaddrs_pase
 #endif
 
 #ifdef HAVE_NETPACKET_PACKET_H
@@ -207,12 +215,6 @@
  * =================================================================== */
 
 
-#ifdef HAVE_SOCKLEN_T
-#  define SOCKLEN_T socklen_t
-#else
-#  define SOCKLEN_T size_t
-#endif
-
 /* Debug stuff... */
 #define NET_NIF_DEBUG_DEFAULT FALSE
 
@@ -278,22 +280,42 @@ ENET_NIF_FUNCS
 /* And here comes the functions that does the actual work (for the most part) */
 static ERL_NIF_TERM enet_command(ErlNifEnv*   env,
                                  ERL_NIF_TERM cmd);
+#if defined(HAVE_GETHOSTNAME)
 static ERL_NIF_TERM enet_gethostname(ErlNifEnv* env);
+#endif
+
+#if defined(HAVE_GETNAMEINFO)
 static ERL_NIF_TERM enet_getnameinfo(ErlNifEnv*          env,
                                      const ESockAddress* saP,
                                      SOCKLEN_T           saLen,
                                      int                 flags);
+#endif
+
+#if defined(HAVE_GETADDRINFO)
 static ERL_NIF_TERM enet_getaddrinfo(ErlNifEnv* env,
                                      char*      host,
                                      char*      serv);
+#endif
+
+#if defined(HAVE_GETIFADDRS) || defined(__PASE__)
 static ERL_NIF_TERM enet_getifaddrs(ErlNifEnv* env,
                                     char*      netns);
+#endif
+
+#if defined(HAVE_IF_NAMETOINDEX)
 static ERL_NIF_TERM enet_if_name2index(ErlNifEnv* env,
                                        char*      ifn);
+#endif
+
+#if defined(HAVE_IF_INDEXTONAME)
 static ERL_NIF_TERM enet_if_index2name(ErlNifEnv*   env,
                                        unsigned int id);
+#endif
+
+#if defined(HAVE_IF_NAMEINDEX) && defined(HAVE_IF_FREENAMEINDEX)
 static ERL_NIF_TERM enet_if_names(ErlNifEnv* env);
 static unsigned int enet_if_names_length(struct if_nameindex* p);
+#endif
 
 /*
 static void net_dtor(ErlNifEnv* env, void* obj);
@@ -312,24 +334,24 @@ static ERL_NIF_TERM enet_getifaddrs(ErlNifEnv* env,
 static ERL_NIF_TERM enet_getifaddrs_process(ErlNifEnv*      env,
                                             struct ifaddrs* ifap);
 static unsigned int enet_getifaddrs_length(struct ifaddrs* ifap);
-static BOOLEAN_T encode_ifaddrs(ErlNifEnv*      env,
-                                struct ifaddrs* ifap,
-                                ERL_NIF_TERM*   eifa);
+static void encode_ifaddrs(ErlNifEnv*      env,
+                           struct ifaddrs* ifap,
+                           ERL_NIF_TERM*   eifa);
 static ERL_NIF_TERM encode_ifaddrs_name(ErlNifEnv* env,
                                         char*      name);
 static ERL_NIF_TERM encode_ifaddrs_flags(ErlNifEnv*   env,
                                          unsigned int flags);
 static ERL_NIF_TERM encode_ifaddrs_addr(ErlNifEnv*       env,
                                         struct sockaddr* sa);
-static char* make_ifaddrs(ErlNifEnv*    env,
-                          ERL_NIF_TERM  name,
-                          ERL_NIF_TERM  flags,
-                          ERL_NIF_TERM  addr,
-                          ERL_NIF_TERM  netmask,
-                          ERL_NIF_TERM  ifu_key,
-                          ERL_NIF_TERM  ifu_value,
-                          ERL_NIF_TERM  data,
-                          ERL_NIF_TERM* ifAddrs);
+static void make_ifaddrs(ErlNifEnv*    env,
+                         ERL_NIF_TERM  name,
+                         ERL_NIF_TERM  flags,
+                         ERL_NIF_TERM  addr,
+                         ERL_NIF_TERM  netmask,
+                         ERL_NIF_TERM  ifu_key,
+                         ERL_NIF_TERM  ifu_value,
+                         ERL_NIF_TERM  data,
+                         ERL_NIF_TERM* ifAddrs);
 #ifdef HAVE_SETNS
 static BOOLEAN_T enet_getifaddrs_netns(ErlNifEnv*   env,
                                        ERL_NIF_TERM map,
@@ -348,8 +370,8 @@ BOOLEAN_T decode_addrinfo_string(ErlNifEnv*         env,
                                  const ERL_NIF_TERM eString,
                                  char**             stringP);
 static ERL_NIF_TERM decode_bool(ErlNifEnv*   env,
-                                ERL_NIF_TERM eBool,
-                                BOOLEAN_T*   bool);
+                                ERL_NIF_TERM ebool,
+                                BOOLEAN_T*   ibool);
 static ERL_NIF_TERM encode_address_infos(ErlNifEnv*       env,
                                          struct addrinfo* addrInfo);
 static ERL_NIF_TERM encode_address_info(ErlNifEnv*       env,
@@ -360,18 +382,16 @@ static ERL_NIF_TERM encode_address_info_family(ErlNifEnv* env,
                                          int        family);
 static ERL_NIF_TERM encode_address_info_type(ErlNifEnv* env,
                                        int        socktype);
-static ERL_NIF_TERM encode_address_info_proto(ErlNifEnv* env,
-                                        int        proto);
 
-static char* make_address_info(ErlNifEnv*    env,
-                               ERL_NIF_TERM  fam,
-                               ERL_NIF_TERM  sockType,
-                               ERL_NIF_TERM  proto,
-                               ERL_NIF_TERM  addr,
-                               ERL_NIF_TERM* ai);
+static void make_address_info(ErlNifEnv*    env,
+                              ERL_NIF_TERM  fam,
+                              ERL_NIF_TERM  sockType,
+                              ERL_NIF_TERM  proto,
+                              ERL_NIF_TERM  addr,
+                              ERL_NIF_TERM* ai);
 
-static BOOLEAN_T extract_debug(ErlNifEnv*   env,
-                               ERL_NIF_TERM map);
+static BOOLEAN_T get_debug(ErlNifEnv*   env,
+                           ERL_NIF_TERM map);
 static int on_load(ErlNifEnv* env, void** priv_data, ERL_NIF_TERM load_info);
 
 
@@ -418,7 +438,6 @@ static const struct in6_addr in6addr_loopback =
     LOCAL_ATOM_DECL(idn);                       \
     LOCAL_ATOM_DECL(master);                    \
     LOCAL_ATOM_DECL(multicast);                 \
-    LOCAL_ATOM_DECL(name);                      \
     LOCAL_ATOM_DECL(namereqd);                  \
     LOCAL_ATOM_DECL(name_info);                 \
     LOCAL_ATOM_DECL(netmask);                   \
@@ -609,7 +628,7 @@ ERL_NIF_TERM nif_gethostname(ErlNifEnv*         env,
 {
 #if defined(__WIN32__)
     return enif_raise_exception(env, MKA(env, "notsup"));
-#else
+#elif defined(HAVE_GETHOSTNAME)
     ERL_NIF_TERM result;
     
     NDBG( ("NET", "nif_gethostname -> entry (%d)\r\n", argc) );
@@ -622,11 +641,13 @@ ERL_NIF_TERM nif_gethostname(ErlNifEnv*         env,
     NDBG( ("NET", "nif_gethostname -> done when result: %T\r\n", result) );
 
     return result;
+#else
+    return esock_make_error(env, esock_atom_enotsup);
 #endif
 }
 
 
-#if !defined(__WIN32__)
+#if !defined(__WIN32__) && defined(HAVE_GETHOSTNAME)
 static
 ERL_NIF_TERM enet_gethostname(ErlNifEnv* env)
 {
@@ -685,13 +706,12 @@ ERL_NIF_TERM nif_getnameinfo(ErlNifEnv*         env,
 {
 #if defined(__WIN32__)
     return enif_raise_exception(env, MKA(env, "notsup"));
-#else
+#elif defined(HAVE_GETNAMEINFO)
     ERL_NIF_TERM result;
     ERL_NIF_TERM eSockAddr, eFlags;
     int          flags = 0; // Just in case...
     ESockAddress sa;
     SOCKLEN_T    saLen = 0; // Just in case...
-    char*        xres;
 
     NDBG( ("NET", "nif_getnameinfo -> entry (%d)\r\n", argc) );
 
@@ -706,9 +726,9 @@ ERL_NIF_TERM nif_getnameinfo(ErlNifEnv*         env,
            "\r\n   Flags:    %T"
            "\r\n", eSockAddr, eFlags) );
 
-    if ((xres = esock_decode_sockaddr(env, eSockAddr, &sa, &saLen)) != NULL) {
-        NDBG( ("NET", "nif_getnameinfo -> failed decode sockaddr: %s\r\n", xres) );
-        return esock_make_error_str(env, xres);
+    if (! esock_decode_sockaddr(env, eSockAddr, &sa, &saLen)) {
+        NDBG( ("NET", "nif_getnameinfo -> failed decode sockaddr\r\n") );
+        return esock_make_error(env, esock_atom_einval);
     }
 
     NDBG( ("NET", "nif_getnameinfo -> (try) decode flags\r\n") );
@@ -723,6 +743,8 @@ ERL_NIF_TERM nif_getnameinfo(ErlNifEnv*         env,
            "\r\n   %T\r\n", result) );
 
     return result;
+#else
+    return esock_make_error(env, esock_atom_enotsup);
 #endif
 }
 
@@ -731,7 +753,7 @@ ERL_NIF_TERM nif_getnameinfo(ErlNifEnv*         env,
 /* Given the provided sock(et) address (and flags), retreive the host and
  * service info.
  */
-#if !defined(__WIN32__)
+#if !defined(__WIN32__) && defined(HAVE_GETNAMEINFO)
 static
 ERL_NIF_TERM enet_getnameinfo(ErlNifEnv*          env,
                               const ESockAddress* saP,
@@ -754,16 +776,13 @@ ERL_NIF_TERM enet_getnameinfo(ErlNifEnv*          env,
     switch (res) {
     case 0:
         {
-            ERL_NIF_TERM keys[] = {atom_host,      atom_service};
-            ERL_NIF_TERM vals[] = {MKS(env, host), MKS(env, serv)};
+            ERL_NIF_TERM keys[]  = {atom_host,      atom_service};
+            ERL_NIF_TERM vals[]  = {MKS(env, host), MKS(env, serv)};
+            size_t       numKeys = NUM(keys);
             ERL_NIF_TERM info;
-            unsigned int numKeys = sizeof(keys) / sizeof(ERL_NIF_TERM);
-            unsigned int numVals = sizeof(vals) / sizeof(ERL_NIF_TERM);
 
-            ESOCK_ASSERT( (numKeys == numVals) );
-
-            if (!MKMA(env, keys, vals, numKeys, &info))
-                return enif_make_badarg(env);
+            ESOCK_ASSERT( numKeys == NUM(vals) );
+            ESOCK_ASSERT( MKMA(env, keys, vals, numKeys, &info) );
 
             result = esock_make_ok2(env, info);
         }
@@ -847,7 +866,7 @@ ERL_NIF_TERM nif_getaddrinfo(ErlNifEnv*         env,
 {
 #if defined(__WIN32__)
     return enif_raise_exception(env, MKA(env, "notsup"));
-#else
+#elif defined(HAVE_GETADDRINFO)
     ERL_NIF_TERM     result, eHostName, eServName; //, eHints;
     char*            hostName;
     char*            servName;
@@ -901,11 +920,13 @@ ERL_NIF_TERM nif_getaddrinfo(ErlNifEnv*         env,
            "\r\n   %T\r\n", result) );
 
     return result;
+#else
+    return esock_make_error(env, esock_atom_enotsup);
 #endif
 }
 
 
-#if !defined(__WIN32__)
+#if !defined(__WIN32__) && defined(HAVE_GETADDRINFO)
 static
 ERL_NIF_TERM enet_getaddrinfo(ErlNifEnv* env,
                               char*      host,
@@ -1030,8 +1051,10 @@ ERL_NIF_TERM nif_getifaddrs(ErlNifEnv*         env,
 {
 #if defined(__WIN32__)
     return enif_raise_exception(env, MKA(env, "notsup"));
-#elif defined(HAVE_GETIFADDRS)
+#elif defined(HAVE_GETIFADDRS) || defined(__PASE__)
+#ifdef HAVE_SETNS
     ERL_NIF_TERM extra;
+#endif
     char*        netns;
     ERL_NIF_TERM result;
 
@@ -1041,7 +1064,10 @@ ERL_NIF_TERM nif_getifaddrs(ErlNifEnv*         env,
         !IS_MAP(env,  argv[0])) {
         return enif_make_badarg(env);
     }
+#ifdef HAVE_SETNS
     extra = argv[0];
+#endif
+
 
 #ifdef HAVE_SETNS
     /* We *currently* only support one extra option: netns */
@@ -1061,12 +1087,12 @@ ERL_NIF_TERM nif_getifaddrs(ErlNifEnv*         env,
 
     return result;
 #else // HAVE_GETIFADDRS
-    return esock_make_error(env, esock_atom_notsup);
+    return esock_make_error(env, esock_atom_enotsup);
 #endif
 }
 
 
-#ifdef HAVE_GETIFADDRS
+#if defined(HAVE_GETIFADDRS) || defined(__PASE__)
 #ifdef HAVE_SETNS
 /* enet_getifaddrs_netns - extract the netns field from the 'extra' map
  *
@@ -1146,9 +1172,17 @@ ERL_NIF_TERM enet_getifaddrs(ErlNifEnv* env, char* netns)
         return esock_make_error_errno(env, save_errno);
 #endif
 
+#ifdef __PASE__
+    if (0 == Qp2getifaddrs(&ifap)) {
+#else
     if (0 == getifaddrs(&ifap)) {
+#endif
         result = enet_getifaddrs_process(env, ifap);
+#ifdef __PASE__
+        Qp2freeifaddrs(ifap);
+#else
         freeifaddrs(ifap);
+#endif
     } else {
         save_errno = get_errno();
 
@@ -1184,18 +1218,14 @@ ERL_NIF_TERM enet_getifaddrs_process(ErlNifEnv* env, struct ifaddrs* ifap)
     NDBG( ("NET", "enet_getifaddrs_process -> len: %d\r\n", len) );
 
     if (len > 0) {
-        ERL_NIF_TERM*   array = MALLOC(len * sizeof(ERL_NIF_TERM));
-        unsigned int    i     = 0;
-        struct ifaddrs* p     = ifap;
+      ERL_NIF_TERM*   array = MALLOC(len * sizeof(ERL_NIF_TERM));
+      unsigned int    i     = 0;
+      struct ifaddrs* p     = ifap;
 
         while (i < len) {
             ERL_NIF_TERM entry;
 
-            if (!encode_ifaddrs(env, p, &entry)) {
-                /* If this fail, we are f*ed, so give up */
-                FREE(array);
-                return esock_make_error(env, esock_atom_einval);
-            }
+            encode_ifaddrs(env, p, &entry);
 
             NDBG( ("NET", "enet_getifaddrs_process -> entry: %T\r\n", entry) );
 
@@ -1249,11 +1279,10 @@ unsigned int enet_getifaddrs_length(struct ifaddrs* ifap)
 
 
 static
-BOOLEAN_T encode_ifaddrs(ErlNifEnv*      env,
-                         struct ifaddrs* ifap,
-                         ERL_NIF_TERM*   eifa)
+void encode_ifaddrs(ErlNifEnv*      env,
+                    struct ifaddrs* ifap,
+                    ERL_NIF_TERM*   eifa)
 {
-    char*        xres;
     ERL_NIF_TERM ename, eflags, eaddr, enetmask, eifu_key, eifu_value, edata;
     ERL_NIF_TERM eifAddrs;
 
@@ -1284,38 +1313,13 @@ BOOLEAN_T encode_ifaddrs(ErlNifEnv*      env,
      */
     edata = esock_atom_undefined;
 
-    if ((xres = make_ifaddrs(env,
-                             ename, eflags, eaddr, enetmask,
-                             eifu_key, eifu_value, edata,
-                             &eifAddrs)) == NULL) {
-        NDBG( ("NET", "encode_ifaddrs -> encoded ifAddrs: %T\r\n", eifAddrs) );
-        *eifa = eifAddrs;
+    make_ifaddrs(env,
+                 ename, eflags, eaddr, enetmask,
+                 eifu_key, eifu_value, edata,
+                 &eifAddrs);
 
-        return TRUE;
-
-    } else {
-
-        /* *** Ouch ***
-         * 
-         * Failed to construct the ifaddrs map => try for a minumum map 
-         * 
-         */
-        ERL_NIF_TERM keys[]  = {atom_name};
-        ERL_NIF_TERM vals[]  = {ename};
-        unsigned int numKeys = sizeof(keys) / sizeof(ERL_NIF_TERM);
-
-        esock_warning_msg("Failed constructing IfAddr map for %s: %s",
-                          ifap->ifa_name, xres);
-
-        if (!MKMA(env, keys, vals, numKeys, eifa)) {
-            // We are in a bad way, give up */
-            *eifa = esock_atom_undefined;
-            return FALSE;
-        } else {
-            return TRUE;
-        }  
-
-    }
+    NDBG( ("NET", "encode_ifaddrs -> encoded ifAddrs: %T\r\n", eifAddrs) );
+    *eifa = eifAddrs;
 }
 
 
@@ -1323,7 +1327,7 @@ BOOLEAN_T encode_ifaddrs(ErlNifEnv*      env,
 static
 ERL_NIF_TERM encode_ifaddrs_name(ErlNifEnv* env, char* name)
 {
-    return ((name == NULL) ? esock_atom_undefined : MKS(env, name));
+  return ((name == NULL) ? esock_atom_undefined : MKS(env, name));
 }
 
 
@@ -1421,12 +1425,11 @@ ERL_NIF_TERM encode_ifaddrs_addr(ErlNifEnv* env, struct sockaddr* sa)
     ERL_NIF_TERM esa;
 
     if (sa != NULL) {
+        
         unsigned int sz = sizeof(ESockAddress);
-        if (esock_encode_sockaddr(env, (ESockAddress*) sa, sz, &esa) != NULL) {
-            NDBG( ("NET", "encode_ifaddrs_addr -> "
-                   "encoding address field 0x%lX failed => use default\r\n", sa) );
-            esa = esock_atom_undefined;
-        }
+
+        esock_encode_sockaddr(env, (ESockAddress*) sa, sz, &esa);
+        
     } else {
         esa = esock_atom_undefined;
     }
@@ -1436,27 +1439,27 @@ ERL_NIF_TERM encode_ifaddrs_addr(ErlNifEnv* env, struct sockaddr* sa)
 
 
 static
-char* make_ifaddrs(ErlNifEnv*    env,
-                   ERL_NIF_TERM  ename,
-                   ERL_NIF_TERM  eflags,
-                   ERL_NIF_TERM  eaddr,
-                   ERL_NIF_TERM  enetmask,
-                   ERL_NIF_TERM  eifu_key,
-                   ERL_NIF_TERM  eifu_value,
-                   ERL_NIF_TERM  edata,
-                   ERL_NIF_TERM* eifAddrs)
+void make_ifaddrs(ErlNifEnv*    env,
+                  ERL_NIF_TERM  ename,
+                  ERL_NIF_TERM  eflags,
+                  ERL_NIF_TERM  eaddr,
+                  ERL_NIF_TERM  enetmask,
+                  ERL_NIF_TERM  eifu_key,
+                  ERL_NIF_TERM  eifu_value,
+                  ERL_NIF_TERM  edata,
+                  ERL_NIF_TERM* eifAddrs)
 {
     /* Several of these values can be (the atom) undefined, which
      * means that they should *not* be included in the result map.
      */
     ERL_NIF_TERM keys[6]; // There are only (at most) siz (6) fields...
     ERL_NIF_TERM vals[6];
-    unsigned int len = sizeof(keys) / sizeof(ERL_NIF_TERM); // Just in case...
-    unsigned int idx = 0;
+    size_t       len = NUM(keys); // Just in case...
+    size_t       idx = 0;
 
     /* *** Name *** */
     NDBG( ("NET", "make_ifaddrs -> name: %T\r\n", ename) );
-    keys[idx] = atom_name;
+    keys[idx] = esock_atom_name;
     vals[idx] = ename;
     idx++;
 
@@ -1511,14 +1514,8 @@ char* make_ifaddrs(ErlNifEnv*    env,
            "\r\n   len: %d"
            "\r\n"
            ) );
-    if (!MKMA(env, keys, vals, len, eifAddrs)) {
-        NDBG( ("NET", "make_ifaddrs -> "
-               "failed contructing interface adress map\r\n") );
-        *eifAddrs = esock_atom_undefined;
-        return ESOCK_STR_EINVAL;
-    } else {
-        return NULL;
-    }
+
+    ESOCK_ASSERT( MKMA(env, keys, vals, len, eifAddrs) );
 }
 
 #endif // HAVE_GETIFADDRS
@@ -1541,7 +1538,7 @@ ERL_NIF_TERM nif_if_name2index(ErlNifEnv*         env,
 {
 #if defined(__WIN32__)
     return enif_raise_exception(env, MKA(env, "notsup"));
-#else
+#elif defined(HAVE_IF_NAMETOINDEX)
     ERL_NIF_TERM eifn, result;
     char         ifn[IF_NAMESIZE+1];
 
@@ -1565,12 +1562,14 @@ ERL_NIF_TERM nif_if_name2index(ErlNifEnv*         env,
     NDBG( ("NET", "nif_if_name2index -> done when result: %T\r\n", result) );
 
     return result;
+#else
+    return esock_make_error(env, esock_atom_enotsup);
 #endif
 }
 
 
 
-#if !defined(__WIN32__)
+#if !defined(__WIN32__) && defined(HAVE_IF_NAMETOINDEX)
 static
 ERL_NIF_TERM enet_if_name2index(ErlNifEnv* env,
                             char*      ifn)
@@ -1613,7 +1612,7 @@ ERL_NIF_TERM nif_if_index2name(ErlNifEnv*         env,
 {
 #if defined(__WIN32__)
     return enif_raise_exception(env, MKA(env, "notsup"));
-#else
+#elif defined(HAVE_IF_INDEXTONAME)
     ERL_NIF_TERM result;
     unsigned int idx;
 
@@ -1633,12 +1632,14 @@ ERL_NIF_TERM nif_if_index2name(ErlNifEnv*         env,
     NDBG( ("NET", "nif_if_index2name -> done when result: %T\r\n", result) );
 
     return result;
+#else
+    return esock_make_error(env, esock_atom_enotsup);
 #endif
 }
 
 
 
-#if !defined(__WIN32__)
+#if !defined(__WIN32__) && defined(HAVE_IF_INDEXTONAME)
 static
 ERL_NIF_TERM enet_if_index2name(ErlNifEnv*   env,
                             unsigned int idx)
@@ -1676,9 +1677,9 @@ ERL_NIF_TERM nif_if_names(ErlNifEnv*         env,
                           int                argc,
                           const ERL_NIF_TERM argv[])
 {
-#if defined(__WIN32__)
+#if defined(__WIN32__) || (defined(__ANDROID__) && (__ANDROID_API__ < 24))
     return enif_raise_exception(env, MKA(env, "notsup"));
-#else
+#elif defined(HAVE_IF_NAMEINDEX) && defined(HAVE_IF_FREENAMEINDEX)
     ERL_NIF_TERM result;
 
     NDBG( ("NET", "nif_if_names -> entry (%d)\r\n", argc) );
@@ -1692,12 +1693,19 @@ ERL_NIF_TERM nif_if_names(ErlNifEnv*         env,
     NDBG( ("NET", "nif_if_names -> done when result: %T\r\n", result) );
 
     return result;
+#else
+    return esock_make_error(env, esock_atom_enotsup);
 #endif
 }
 
 
 
-#if !defined(__WIN32__)
+/* if_nameindex and if_freenameindex were added in Android 7.0 Nougat. With
+the Android NDK Unified Headers, check that the build is targeting at least
+the corresponding API level 24. */
+/* Can we replace the ANDROID tests with the HAVE_... ? */
+#if !defined(__WIN32__) && !(defined(__ANDROID__) && (__ANDROID_API__ < 24))
+#if defined(HAVE_IF_NAMEINDEX) && defined(HAVE_IF_FREENAMEINDEX)
 static
 ERL_NIF_TERM enet_if_names(ErlNifEnv* env)
 {
@@ -1770,7 +1778,8 @@ unsigned int enet_if_names_length(struct if_nameindex* p)
 
     return len;
 }
-#endif // if !defined(__WIN32__)
+#endif // if defined(HAVE_IF_NAMEINDEX) && defined(HAVE_IF_FREENAMEINDEX)
+#endif // if !defined(__WIN32__) && ...
 
 
 
@@ -1931,14 +1940,14 @@ BOOLEAN_T decode_addrinfo_string(ErlNifEnv*         env,
 
 static
 ERL_NIF_TERM decode_bool(ErlNifEnv*   env,
-                         ERL_NIF_TERM eBool,
-                         BOOLEAN_T*   bool)
+                         ERL_NIF_TERM ebool,
+                         BOOLEAN_T*   ibool)
 {
-    if (COMPARE(eBool, esock_atom_true) == 0) {
-        *bool = TRUE;
+    if (COMPARE(ebool, esock_atom_true) == 0) {
+        *ibool = TRUE;
         return esock_atom_ok;
-    } else if (COMPARE(eBool, esock_atom_false) == 0) {
-        *bool = FALSE;
+    } else if (COMPARE(ebool, esock_atom_false) == 0) {
+        *ibool = FALSE;
         return esock_atom_ok;
     } else {
         return esock_make_error(env, esock_atom_einval);
@@ -2025,17 +2034,14 @@ ERL_NIF_TERM encode_address_info(ErlNifEnv*       env,
 
     fam   = encode_address_info_family(env, addrInfoP->ai_family);
     type  = encode_address_info_type(env,   addrInfoP->ai_socktype);
-    proto = encode_address_info_proto(env,  addrInfoP->ai_protocol);
+    proto = MKI(env, addrInfoP->ai_protocol);
     esock_encode_sockaddr(env,
                           (ESockAddress*) addrInfoP->ai_addr,
                           addrInfoP->ai_addrlen,
                           &addr);
     
-    if (make_address_info(env, fam, type, proto, addr, &addrInfo) == NULL)
-        return addrInfo;
-    else
-        return esock_atom_undefined; // We should do better...
-
+    make_address_info(env, fam, type, proto, addr, &addrInfo);
+    return addrInfo;
 }
 
 
@@ -2050,9 +2056,7 @@ ERL_NIF_TERM encode_address_info_family(ErlNifEnv* env,
 {
     ERL_NIF_TERM efam;
 
-    if (NULL != esock_encode_domain(env, family, &efam))
-        efam = MKI(env, family);
-
+    esock_encode_domain(env, family, &efam);
     return efam;
 }
 
@@ -2069,57 +2073,29 @@ ERL_NIF_TERM encode_address_info_type(ErlNifEnv* env,
 {
     ERL_NIF_TERM etype;
 
-    if (NULL != esock_encode_type(env, socktype, &etype))
-        etype = MKI(env, socktype);
-
+    esock_encode_type(env, socktype, &etype);
     return etype;
 }
 
 
 
-/* Convert an "native" protocol to an erlang protocol.
- * Note that this is not currently exhaustive, but only supports
- * tcp and udp. Other values will be returned as is, that is
- * in the form of an integer.
- */
 static
-ERL_NIF_TERM encode_address_info_proto(ErlNifEnv* env,
-                                       int        proto)
+void make_address_info(ErlNifEnv*    env,
+                       ERL_NIF_TERM  fam,
+                       ERL_NIF_TERM  sockType,
+                       ERL_NIF_TERM  proto,
+                       ERL_NIF_TERM  addr,
+                       ERL_NIF_TERM* ai)
 {
-    ERL_NIF_TERM eproto;
-
-    if (NULL != esock_encode_protocol(env, proto, &eproto))
-        eproto = MKI(env, proto);
-
-    return eproto;
-}
-
-
-
-static
-char* make_address_info(ErlNifEnv*    env,
-                        ERL_NIF_TERM  fam,
-                        ERL_NIF_TERM  sockType,
-                        ERL_NIF_TERM  proto,
-                        ERL_NIF_TERM  addr,
-                        ERL_NIF_TERM* ai)
-{
-    ERL_NIF_TERM keys[] = {esock_atom_family,
-                           esock_atom_type,
-                           esock_atom_protocol,
-                           esock_atom_addr};
-    ERL_NIF_TERM vals[] = {fam, sockType, proto, addr};
-    unsigned int numKeys = sizeof(keys) / sizeof(ERL_NIF_TERM);
-    unsigned int numVals = sizeof(vals) / sizeof(ERL_NIF_TERM);
+    ERL_NIF_TERM keys[]  = {esock_atom_family,
+                            esock_atom_type,
+                            esock_atom_protocol,
+                            esock_atom_addr};
+    ERL_NIF_TERM vals[]  = {fam, sockType, proto, addr};
+    size_t       numKeys = NUM(keys);
     
-    ESOCK_ASSERT( (numKeys == numVals) );
-    
-    if (!MKMA(env, keys, vals, numKeys, ai)) {
-        *ai = esock_atom_undefined;
-        return ESOCK_STR_EINVAL;
-    } else {
-        return NULL;
-    }
+    ESOCK_ASSERT( numKeys == NUM(vals) );
+    ESOCK_ASSERT( MKMA(env, keys, vals, numKeys, ai) );
 }
 #endif // if !defined(__WIN32__)
 
@@ -2302,8 +2278,8 @@ ErlNifFunc net_funcs[] =
 
 #if !defined(__WIN32__)
 static
-BOOLEAN_T extract_debug(ErlNifEnv*   env,
-                        ERL_NIF_TERM map)
+BOOLEAN_T get_debug(ErlNifEnv*   env,
+                    ERL_NIF_TERM map)
 {
     /*
      * We need to do this here since the "proper" atom has not been
@@ -2311,7 +2287,7 @@ BOOLEAN_T extract_debug(ErlNifEnv*   env,
      */
     ERL_NIF_TERM debug = MKA(env, "debug");
     
-    return esock_extract_bool_from_map(env, map, debug, NET_NIF_DEBUG_DEFAULT);
+    return esock_get_bool_from_map(env, map, debug, NET_NIF_DEBUG_DEFAULT);
 }
 #endif
 
@@ -2325,7 +2301,7 @@ int on_load(ErlNifEnv* env, void** priv_data, ERL_NIF_TERM load_info)
 {
 #if !defined(__WIN32__)    
     // We should make it possible to use load_info to get default values
-    data.debug = extract_debug(env, load_info);
+    data.debug = get_debug(env, load_info);
 
     NDBG( ("NET", "on_load -> entry\r\n") );
 #endif
