@@ -1,7 +1,7 @@
 /*
  * %CopyrightBegin%
  *
- * Copyright Ericsson 2017-2020. All Rights Reserved.
+ * Copyright Ericsson 2017-2022. All Rights Reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -84,9 +84,20 @@
 typedef struct {
     efile_data_t common;
     HANDLE handle;
+    /* The following field is only used when the handle has been
+       obtained from an already exisiting file descriptor (i.e.,
+       prim_file:file_desc_to_ref/2). common.modes is set to
+       EFILE_MODE_FROM_ALREADY_OPEN_FD when that is the case. It is
+       needed because we can't close using handle in that case. */
+    int fd;
 } efile_win_t;
 
 static int windows_to_posix_errno(DWORD last_error);
+static void tmp_nop_invalid_parameter_handler(const wchar_t* expression,
+                                              const wchar_t* function,
+                                              const wchar_t* file,
+                                              unsigned int line,
+                                              uintptr_t pReserved);
 
 static int has_invalid_null_termination(const ErlNifBinary *path) {
     const WCHAR *null_pos, *end_pos;
@@ -207,14 +218,21 @@ posix_errno_t efile_marshal_path(ErlNifEnv *env, ERL_NIF_TERM path, efile_path_t
 
 ERL_NIF_TERM efile_get_handle(ErlNifEnv *env, efile_data_t *d) {
     efile_win_t *w = (efile_win_t*)d;
-
-    ERL_NIF_TERM result;
+    ERL_NIF_TERM handle;
     unsigned char *bits;
 
-    bits = enif_make_new_binary(env, sizeof(w->handle), &result);
+    bits = enif_make_new_binary(env, sizeof(w->handle), &handle);
     memcpy(bits, &w->handle, sizeof(w->handle));
 
-    return result;
+    return handle;
+}
+
+posix_errno_t efile_dup_handle(ErlNifEnv *env, efile_data_t *d, ErlNifEvent *handle) {
+    (void) env;
+    (void) d;
+    (void) handle;
+    /* XXX not implemented yet */
+    return ENOTSUP;
 }
 
 /** @brief Converts a native path to the preferred form in "erlang space,"
@@ -491,10 +509,59 @@ posix_errno_t efile_open(const efile_path_t *path, enum efile_modes_t modes,
     }
 }
 
+static void tmp_nop_invalid_parameter_handler(const wchar_t* expression,
+                                              const wchar_t* function,
+                                              const wchar_t* file,
+                                              unsigned int line,
+                                              uintptr_t pReserved) {
+    (void)expression;
+    (void)function;
+    (void)file;
+    (void)line;
+    (void)pReserved;
+}
+
+posix_errno_t efile_from_fd(int fd,
+                            ErlNifResourceType *nif_type,
+                            efile_data_t **d) {
+    HANDLE handle;
+
+    _invalid_parameter_handler old_handler;
+
+    /* Temporarily disable the parameter handler so we don't crash */
+    old_handler =
+        _set_thread_local_invalid_parameter_handler(tmp_nop_invalid_parameter_handler);
+
+    handle = (HANDLE)_get_osfhandle(fd);
+
+    /* Enable old parameter handler again */
+    _set_thread_local_invalid_parameter_handler(old_handler);
+
+    if(handle != INVALID_HANDLE_VALUE && handle != ((HANDLE)-2)) {
+        efile_win_t *w;
+
+        w = (efile_win_t*)enif_alloc_resource(nif_type, sizeof(efile_win_t));
+        w->handle = handle;
+        w->fd = fd;
+
+        EFILE_INIT_RESOURCE(&w->common, EFILE_MODE_FROM_ALREADY_OPEN_FD);
+        (*d) = &w->common;
+
+        return 0;
+    } else {
+        return EBADF;
+    }
+}
+
 int efile_close(efile_data_t *d, posix_errno_t *error) {
     efile_win_t *w = (efile_win_t*)d;
     HANDLE handle;
-
+    int from_already_open_fd =
+        w->common.modes & EFILE_MODE_FROM_ALREADY_OPEN_FD;
+    int fd;
+    if (from_already_open_fd) {
+        fd = w->fd;
+    }
     ASSERT(enif_thread_type() == ERL_NIF_THR_DIRTY_IO_SCHEDULER);
     ASSERT(erts_atomic32_read_nob(&d->state) == EFILE_STATE_CLOSED);
     ASSERT(w->handle != INVALID_HANDLE_VALUE);
@@ -504,7 +571,12 @@ int efile_close(efile_data_t *d, posix_errno_t *error) {
 
     enif_release_resource(d);
 
-    if(!CloseHandle(handle)) {
+    if(from_already_open_fd) {
+        if(_close(fd) == -1) {
+            *error = EBADF;
+            return 0;
+        }
+    } else if(!CloseHandle(handle)) {
         *error = windows_to_posix_errno(GetLastError());
         return 0;
     }
@@ -558,7 +630,8 @@ static void shift_iov(SysIOVec **iov, int *iovlen, DWORD shift) {
 typedef BOOL (WINAPI *io_op_t)(HANDLE, LPVOID, DWORD, LPDWORD, LPOVERLAPPED);
 
 static Sint64 internal_sync_io(efile_win_t *w, io_op_t operation,
-        SysIOVec *iov, int iovlen, OVERLAPPED *overlapped) {
+                               SysIOVec *iov, int iovlen, OVERLAPPED *overlapped,
+                               int is_read) {
 
     Sint64 bytes_processed = 0;
 
@@ -574,10 +647,17 @@ static Sint64 internal_sync_io(efile_win_t *w, io_op_t operation,
             &block_bytes_processed, overlapped);
         last_error = GetLastError();
 
-        if(!succeeded && (last_error != ERROR_HANDLE_EOF)) {
-            w->common.posix_errno = windows_to_posix_errno(last_error);
-            return -1;
-        } else if(block_bytes_processed == 0) {
+        if(!succeeded) {
+            if(is_read && last_error == ERROR_BROKEN_PIPE) {
+                /* Pipes gives ERROR_BROKEN_PIPE instead of EOF when the
+                   write end has been closed */
+                return bytes_processed;
+            } else if(!is_read || last_error != ERROR_HANDLE_EOF) {
+                w->common.posix_errno = windows_to_posix_errno(last_error);
+                return -1;
+            }
+        }
+        if(block_bytes_processed == 0) {
             /* EOF */
             return bytes_processed;
         }
@@ -595,7 +675,7 @@ static Sint64 internal_sync_io(efile_win_t *w, io_op_t operation,
 Sint64 efile_readv(efile_data_t *d, SysIOVec *iov, int iovlen) {
     efile_win_t *w = (efile_win_t*)d;
 
-    return internal_sync_io(w, ReadFile, iov, iovlen, NULL);
+    return internal_sync_io(w, ReadFile, iov, iovlen, NULL, 1);
 }
 
 Sint64 efile_writev(efile_data_t *d, SysIOVec *iov, int iovlen) {
@@ -614,7 +694,7 @@ Sint64 efile_writev(efile_data_t *d, SysIOVec *iov, int iovlen) {
         overlapped = NULL;
     }
 
-    return internal_sync_io(w, WriteFile, iov, iovlen, overlapped);
+    return internal_sync_io(w, WriteFile, iov, iovlen, overlapped, 0);
 }
 
 Sint64 efile_preadv(efile_data_t *d, Sint64 offset, SysIOVec *iov, int iovlen) {
@@ -626,7 +706,7 @@ Sint64 efile_preadv(efile_data_t *d, Sint64 offset, SysIOVec *iov, int iovlen) {
     overlapped.OffsetHigh = (offset >> 32) & 0xFFFFFFFF;
     overlapped.Offset = offset & 0xFFFFFFFF;
 
-    return internal_sync_io(w, ReadFile, iov, iovlen, &overlapped);
+    return internal_sync_io(w, ReadFile, iov, iovlen, &overlapped, 1);
 }
 
 Sint64 efile_pwritev(efile_data_t *d, Sint64 offset, SysIOVec *iov, int iovlen) {
@@ -638,7 +718,7 @@ Sint64 efile_pwritev(efile_data_t *d, Sint64 offset, SysIOVec *iov, int iovlen)
     overlapped.OffsetHigh = (offset >> 32) & 0xFFFFFFFF;
     overlapped.Offset = offset & 0xFFFFFFFF;
 
-    return internal_sync_io(w, WriteFile, iov, iovlen, &overlapped);
+    return internal_sync_io(w, WriteFile, iov, iovlen, &overlapped, 0);
 }
 
 int efile_seek(efile_data_t *d, enum efile_seek_t seek, Sint64 offset, Sint64 *new_position) {
@@ -893,11 +973,12 @@ posix_errno_t efile_read_info(const efile_path_t *path, int follow_links, efile_
 
             CloseHandle(handle);
 
-            if(posix_errno != 0) {
-                return posix_errno;
+            if(posix_errno == 0) {
+                posix_errno = efile_read_info(&resolved_path, 0, result);
+                enif_release_binary(&resolved_path);
             }
 
-            return efile_read_info(&resolved_path, 0, result);
+            return posix_errno;
         }
 
         GetFileInformationByHandle(handle, &native_file_info);
@@ -928,17 +1009,20 @@ posix_errno_t efile_read_handle_info(efile_data_t *d, efile_fileinfo_t *result)
 
     if(GetFileInformationByHandle(w->handle, &native_file_info)) {
         build_file_info_times(&native_file_info, result);
+        build_file_info(&native_file_info, &path, 0, result);
+
+        posix_errno = 0;
     } else if(is_path_root(&path)) {
         /* GetFileInformationByHandle is not supported on path roots, so
          * fall back to efile_read_info. */
-        return efile_read_info(&path, 0, result);
+        posix_errno = efile_read_info(&path, 0, result);
     } else {
-        return windows_to_posix_errno(GetLastError());
+        posix_errno = windows_to_posix_errno(GetLastError());
     }
 
-    build_file_info(&native_file_info, &path, 0, result);
+    enif_release_binary(&path);
 
-    return 0;
+    return posix_errno;
 }
 
 posix_errno_t efile_set_permissions(const efile_path_t *path, Uint32 permissions) {
